"""Auto-instrumentation — framework adapters and generic decorator.

All instrumentors follow the same protocol:

.. code-block:: python

    from agent_observability.auto_instrument import LangChainInstrumentor
    from agent_observability.tracer import AgentTracer

    tracer = AgentTracer(agent_id="my-agent")
    instr = LangChainInstrumentor(tracer)
    instr.instrument()
    # ... your LangChain code ...
    instr.uninstrument()
"""
from __future__ import annotations

from agent_observability.auto_instrument.anthropic_sdk import AnthropicInstrumentor
from agent_observability.auto_instrument.autogen import AutoGenInstrumentor
from agent_observability.auto_instrument.crewai import CrewAIInstrumentor
from agent_observability.auto_instrument.generic import GenericInstrumentor
from agent_observability.auto_instrument.langchain import LangChainInstrumentor
from agent_observability.auto_instrument.openai_sdk import OpenAIInstrumentor

__all__ = [
    "LangChainInstrumentor",
    "CrewAIInstrumentor",
    "AutoGenInstrumentor",
    "AnthropicInstrumentor",
    "OpenAIInstrumentor",
    "GenericInstrumentor",
]



"""Unit tests for cost.pricing — ModelPricing, PROVIDER_PRICING, get_pricing, estimate_cost."""
from __future__ import annotations

import pytest

from agent_observability.cost.pricing import (
    PROVIDER_PRICING,
    ModelPricing,
    estimate_cost,
    get_pricing,
    register_pricing,
)


class TestModelPricingComputeCost:
    def test_zero_tokens_produce_zero_cost(self) -> None:
        pricing = ModelPricing(
            provider="test", model="m", input_per_million=3.0, output_per_million=15.0
        )
        assert pricing.compute_cost(0, 0) == 0.0

    def test_one_million_input_tokens_costs_input_price(self) -> None:
        pricing = ModelPricing(
            provider="anthropic",
            model="claude-3-5-sonnet-20241022",
            input_per_million=3.0,
            output_per_million=15.0,
        )
        cost = pricing.compute_cost(1_000_000, 0)
        assert cost == pytest.approx(3.0, rel=1e-6)

    def test_one_million_output_tokens_costs_output_price(self) -> None:
        pricing = ModelPricing(
            provider="openai",
            model="gpt-4o",
            input_per_million=2.5,
            output_per_million=10.0,
        )
        cost = pricing.compute_cost(0, 1_000_000)
        assert cost == pytest.approx(10.0, rel=1e-6)

    def test_mixed_tokens_sum_correctly(self) -> None:
        pricing = ModelPricing(
            provider="openai",
            model="gpt-4o",
            input_per_million=2.5,
            output_per_million=10.0,
        )
        # 500k input @ $2.50/M = $1.25; 200k output @ $10/M = $2.00
        cost = pricing.compute_cost(500_000, 200_000)
        assert cost == pytest.approx(3.25, rel=1e-6)

    def test_cached_tokens_charged_at_cached_rate(self) -> None:
        pricing = ModelPricing(
            provider="anthropic",
            model="claude-3-5-sonnet-20241022",
            input_per_million=3.0,
            output_per_million=15.0,
            cached_input_per_million=0.30,
        )
        # 1M cached input @ $0.30/M = $0.30
        cost = pricing.compute_cost(0, 0, cached_input_tokens=1_000_000)
        assert cost == pytest.approx(0.30, rel=1e-6)

    def test_result_rounded_to_8_decimal_places(self) -> None:
        pricing = ModelPricing(
            provider="test",
            model="m",
            input_per_million=1.0,
            output_per_million=1.0,
        )
        cost = pricing.compute_cost(1, 0)
        # Should be a nice rounded float
        assert isinstance(cost, float)

    def test_gpt4o_mini_low_cost_calculation(self) -> None:
        pricing = ModelPricing(
            provider="openai",
            model="gpt-4o-mini",
            input_per_million=0.15,
            output_per_million=0.60,
        )
        # 10k input, 5k output
        cost = pricing.compute_cost(10_000, 5_000)
        expected = (10_000 / 1_000_000) * 0.15 + (5_000 / 1_000_000) * 0.60
        assert cost == pytest.approx(expected, rel=1e-6)


class TestProviderPricingRegistry:
    def test_anthropic_claude_sonnet_in_registry(self) -> None:
        assert "anthropic/claude-3-5-sonnet-20241022" in PROVIDER_PRICING

    def test_openai_gpt4o_in_registry(self) -> None:
        assert "openai/gpt-4o" in PROVIDER_PRICING

    def test_google_gemini_in_registry(self) -> None:
        assert "google/gemini-1.5-pro" in PROVIDER_PRICING

    def test_all_entries_have_positive_prices(self) -> None:
        for key, pricing in PROVIDER_PRICING.items():
            assert pricing.input_per_million >= 0, f"{key} has negative input price"
            assert pricing.output_per_million >= 0, f"{key} has negative output price"

    def test_all_entries_have_provider_and_model_matching_key(self) -> None:
        for key, pricing in PROVIDER_PRICING.items():
            expected_key = f"{pricing.provider}/{pricing.model}"
            assert expected_key == key


class TestGetPricing:
    def test_exact_lookup_returns_pricing(self) -> None:
        result = get_pricing("openai", "gpt-4o")
        assert result is not None
        assert result.model == "gpt-4o"

    def test_missing_model_returns_none(self) -> None:
        result = get_pricing("openai", "gpt-99-turbo-ultra")
        assert result is None

    def test_prefix_match_resolves_partial_model_name(self) -> None:
        result = get_pricing("anthropic", "claude-3-5-sonnet")
        assert result is not None

    def test_model_only_lookup_as_last_resort(self) -> None:
        result = get_pricing("unknown_provider", "gpt-4o")
        assert result is not None
        assert result.model == "gpt-4o"

    def test_anthropic_haiku_returns_correct_prices(self) -> None:
        result = get_pricing("anthropic", "claude-3-5-haiku-20241022")
        assert result is not None
        assert result.input_per_million == pytest.approx(0.80)
        assert result.output_per_million == pytest.approx(4.00)


class TestRegisterPricing:
    def test_register_new_model_makes_it_discoverable(self) -> None:
        custom = ModelPricing(
            provider="custom_corp",
            model="my-proprietary-model",
            input_per_million=5.0,
            output_per_million=20.0,
        )
        register_pricing(custom)
        found = get_pricing("custom_corp", "my-proprietary-model")
        assert found is not None
        assert found.input_per_million == 5.0

    def test_register_overrides_existing_entry(self) -> None:
        override = ModelPricing(
            provider="openai",
            model="gpt-4o-override",
            input_per_million=999.0,
            output_per_million=999.0,
        )
        register_pricing(override)
        found = get_pricing("openai", "gpt-4o-override")
        assert found is not None
        assert found.input_per_million == 999.0


class TestEstimateCost:
    def test_known_model_returns_positive_cost(self) -> None:
        cost = estimate_cost("openai", "gpt-4o", 1000, 500)
        assert cost > 0.0

    def test_unknown_model_returns_zero(self) -> None:
        cost = estimate_cost("nonexistent", "no-model", 1000, 500)
        assert cost == 0.0

    def test_zero_tokens_returns_zero(self) -> None:
        cost = estimate_cost("openai", "gpt-4o", 0, 0)
        assert cost == 0.0



"""AnthropicInstrumentor — auto-trace the Anthropic Python SDK via monkey-patching.

Wraps ``anthropic.Anthropic.messages.create`` (and the async variant) so that
every Claude API call is captured as an ``llm_call`` span.  The Anthropic
import is guarded; if the SDK is not installed the instrumentor is a no-op.
"""
from __future__ import annotations

import logging
from typing import Optional

from agent_observability.tracer.agent_tracer import AgentTracer

logger = logging.getLogger(__name__)

try:
    import anthropic  # noqa: F401

    _ANTHROPIC_AVAILABLE = True
except ImportError:
    _ANTHROPIC_AVAILABLE = False


class AnthropicInstrumentor:
    """Instrument the Anthropic SDK with agent-semantic tracing.

    Parameters
    ----------
    tracer:
        The :class:`~agent_observability.tracer.AgentTracer` instance to use.
    """

    def __init__(self, tracer: AgentTracer) -> None:
        self._tracer = tracer
        self._original_messages_create: Optional[object] = None
        self._instrumented = False

    def instrument(self) -> None:
        """Monkey-patch the Anthropic SDK to add tracing.

        Patches ``anthropic.resources.messages.Messages.create``.
        Safe to call multiple times.
        """
        if self._instrumented:
            return
        if not _ANTHROPIC_AVAILABLE:
            logger.debug("AnthropicInstrumentor: anthropic not installed, skipping")
            return

        tracer = self._tracer

        try:
            from anthropic.resources.messages import Messages

            original_create = Messages.create

            def _traced_create(
                self_messages: object,
                *args: object,
                **kwargs: object,
            ) -> object:
                model = str(kwargs.get("model", "claude-unknown"))
                max_tokens = int(kwargs.get("max_tokens", 0))
                messages_list = kwargs.get("messages", [])
                prompt_token_estimate = sum(
                    len(str(m.get("content", ""))) // 4
                    for m in messages_list
                    if isinstance(m, dict)
                )
                with tracer.llm_call(
                    model=model,
                    provider="anthropic",
                    prompt_tokens=prompt_token_estimate,
                ) as span:
                    response = original_create(self_messages, *args, **kwargs)  # type: ignore[operator]
                    # Extract usage from response if available
                    usage = getattr(response, "usage", None)
                    if usage is not None:
                        input_tok = getattr(usage, "input_tokens", 0)
                        output_tok = getattr(usage, "output_tokens", 0)
                        span.set_tokens(input_tok, output_tok)
                    if max_tokens:
                        span.set_attribute("llm.max_tokens", max_tokens)
                    return response

            Messages.create = _traced_create  # type: ignore[method-assign]
            self._original_messages_create = original_create
            logger.debug("AnthropicInstrumentor: patched Messages.create")
        except (ImportError, AttributeError) as exc:
            logger.debug("AnthropicInstrumentor: could not patch Messages.create: %s", exc)

        self._instrumented = True

    def uninstrument(self) -> None:
        """Remove all patches installed by :meth:`instrument`."""
        if not self._instrumented:
            return

        try:
            from anthropic.resources.messages import Messages

            if self._original_messages_create is not None:
                Messages.create = self._original_messages_create  # type: ignore[method-assign]
                self._original_messages_create = None
        except (ImportError, AttributeError):
            pass

        self._instrumented = False
        logger.debug("AnthropicInstrumentor: uninstrumented")



"""Tests for tracing.exporter (JsonLinesExporter, ConsoleSpanExporter,
AgentSpanExporter, build_otlp_exporter) and tracing.tracer
(AgentTracerProvider, setup_tracing, _NoOpTracer, _NoOpSpanHandle)."""
from __future__ import annotations

import io
import json
import sys
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock, patch

import pytest

from agent_observability.tracing.exporter import (
    AgentSpanExporter,
    ConsoleSpanExporter,
    JsonLinesExporter,
    build_otlp_exporter,
)
from agent_observability.tracing.tracer import (
    AgentTracerProvider,
    _NoOpSpanHandle,
    _NoOpTracer,
    setup_tracing,
)


# ── Helper ─────────────────────────────────────────────────────────────────────


def _make_mock_span(name: str = "test-span") -> MagicMock:
    """Build a mock ReadableSpan-like object for exporter tests."""
    ctx = MagicMock()
    ctx.trace_id = 0xABCD1234ABCD1234ABCD1234ABCD1234
    ctx.span_id = 0x1234ABCD1234ABCD
    span = MagicMock()
    span.name = name
    span.context = ctx
    span.parent = None
    span.start_time = 1_000_000_000
    span.end_time = 1_100_000_000
    span.status = MagicMock()
    span.status.status_code = "OK"
    span.attributes = {"agent.id": "agent-1"}
    span.events = []
    return span


# ── JsonLinesExporter ──────────────────────────────────────────────────────────


class TestJsonLinesExporter:
    def test_export_to_file(self, tmp_path: Path) -> None:
        output_file = str(tmp_path / "spans.jsonl")
        exporter = JsonLinesExporter(path=output_file)
        mock_span = _make_mock_span("my-span")

        with patch("agent_observability.tracing.exporter._OTEL_SDK_AVAILABLE", True):
            mock_result = MagicMock()
            mock_result.SUCCESS = "SUCCESS"
            with patch("agent_observability.tracing.exporter.SpanExportResult", mock_result):
                exporter.export([mock_span])

        assert Path(output_file).exists()
        content = Path(output_file).read_text(encoding="utf-8")
        assert "my-span" in content

    def test_export_no_otel_returns_none(self) -> None:
        exporter = JsonLinesExporter()
        with patch("agent_observability.tracing.exporter._OTEL_SDK_AVAILABLE", False):
            result = exporter.export([_make_mock_span()])
        assert result is None

    def test_export_to_stdout(self, capsys: Any) -> None:
        exporter = JsonLinesExporter(path="-")
        mock_span = _make_mock_span("stdout-span")

        with patch("agent_observability.tracing.exporter._OTEL_SDK_AVAILABLE", True):
            mock_result = MagicMock()
            mock_result.SUCCESS = "SUCCESS"
            with patch("agent_observability.tracing.exporter.SpanExportResult", mock_result):
                exporter.export([mock_span])

        captured = capsys.readouterr()
        assert "stdout-span" in captured.out

    def test_export_empty_path_is_stdout(self) -> None:
        exporter = JsonLinesExporter(path="")
        assert exporter._to_stdout is True

    def test_shutdown_noop(self) -> None:
        exporter = JsonLinesExporter()
        exporter.shutdown()  # should not raise

    def test_force_flush_returns_true(self) -> None:
        exporter = JsonLinesExporter()
        assert exporter.force_flush() is True

    def test_export_appends_to_existing_file(self, tmp_path: Path) -> None:
        output_file = str(tmp_path / "spans.jsonl")
        Path(output_file).write_text("existing\n", encoding="utf-8")
        exporter = JsonLinesExporter(path=output_file)
        mock_span = _make_mock_span("new-span")

        with patch("agent_observability.tracing.exporter._OTEL_SDK_AVAILABLE", True):
            mock_result = MagicMock()
            mock_result.SUCCESS = "SUCCESS"
            with patch("agent_observability.tracing.exporter.SpanExportResult", mock_result):
                exporter.export([mock_span])

        content = Path(output_file).read_text(encoding="utf-8")
        assert "existing" in content
        assert "new-span" in content


# ── ConsoleSpanExporter ────────────────────────────────────────────────────────


class TestConsoleSpanExporter:
    def test_export_no_otel_returns_none(self) -> None:
        exporter = ConsoleSpanExporter()
        with patch("agent_observability.tracing.exporter._OTEL_SDK_AVAILABLE", False):
            result = exporter.export([_make_mock_span()])
        assert result is None

    def test_export_logs_spans(self) -> None:
        exporter = ConsoleSpanExporter()
        mock_span = _make_mock_span("logged-span")

        with patch("agent_observability.tracing.exporter._OTEL_SDK_AVAILABLE", True):
            mock_result = MagicMock()
            mock_result.SUCCESS = "SUCCESS"
            with patch("agent_observability.tracing.exporter.SpanExportResult", mock_result):
                with patch("agent_observability.tracing.exporter.logger") as mock_logger:
                    exporter.export([mock_span])
                    mock_logger.info.assert_called()

    def test_shutdown_noop(self) -> None:
        exporter = ConsoleSpanExporter()
        exporter.shutdown()  # should not raise

    def test_force_flush_returns_true(self) -> None:
        exporter = ConsoleSpanExporter()
        assert exporter.force_flush() is True


# ── AgentSpanExporter ──────────────────────────────────────────────────────────


class TestAgentSpanExporter:
    def test_default_creates_console_exporter(self) -> None:
        exporter = AgentSpanExporter()
        assert len(exporter._exporters) == 1
        assert isinstance(exporter._exporters[0], ConsoleSpanExporter)

    def test_add_exporter(self) -> None:
        exporter = AgentSpanExporter(exporters=[ConsoleSpanExporter()])
        new_exp = ConsoleSpanExporter()
        exporter.add_exporter(new_exp)
        assert len(exporter._exporters) == 2

    def test_export_no_otel_returns_none(self) -> None:
        exporter = AgentSpanExporter()
        with patch("agent_observability.tracing.exporter._OTEL_SDK_AVAILABLE", False):
            result = exporter.export([_make_mock_span()])
        assert result is None

    def test_export_forwards_to_downstream(self) -> None:
        mock_downstream = MagicMock()
        mock_downstream.export.return_value = "SUCCESS"
        exporter = AgentSpanExporter(exporters=[mock_downstream])

        with patch("agent_observability.tracing.exporter._OTEL_SDK_AVAILABLE", True):
            mock_result = MagicMock()
            mock_result.SUCCESS = "SUCCESS"
            with patch("agent_observability.tracing.exporter.SpanExportResult", mock_result):
                exporter.export([_make_mock_span()])
                mock_downstream.export.assert_called_once()

    def test_export_handles_downstream_exception(self) -> None:
        mock_downstream = MagicMock()
        mock_downstream.export.side_effect = RuntimeError("exporter failed")
        exporter = AgentSpanExporter(exporters=[mock_downstream])

        with patch("agent_observability.tracing.exporter._OTEL_SDK_AVAILABLE", True):
            mock_result = MagicMock()
            mock_result.SUCCESS = "SUCCESS"
            mock_result.FAILURE = "FAILURE"
            with patch("agent_observability.tracing.exporter.SpanExportResult", mock_result):
                result = exporter.export([_make_mock_span()])
                assert result == "FAILURE"

    def test_shutdown_calls_all_exporters(self) -> None:
        mock_a = MagicMock()
        mock_b = MagicMock()
        exporter = AgentSpanExporter(exporters=[mock_a, mock_b])
        exporter.shutdown()
        mock_a.shutdown.assert_called_once()
        mock_b.shutdown.assert_called_once()

    def test_shutdown_handles_exporter_exception(self) -> None:
        mock_exp = MagicMock()
        mock_exp.shutdown.side_effect = RuntimeError("shutdown error")
        exporter = AgentSpanExporter(exporters=[mock_exp])
        exporter.shutdown()  # should not raise

    def test_force_flush_all_success(self) -> None:
        mock_a = MagicMock()
        mock_a.force_flush.return_value = True
        mock_b = MagicMock()
        mock_b.force_flush.return_value = True
        exporter = AgentSpanExporter(exporters=[mock_a, mock_b])
        assert exporter.force_flush() is True

    def test_force_flush_partial_failure(self) -> None:
        mock_a = MagicMock()
        mock_a.force_flush.return_value = True
        mock_b = MagicMock()
        mock_b.force_flush.return_value = False
        exporter = AgentSpanExporter(exporters=[mock_a, mock_b])
        assert exporter.force_flush() is False

    def test_force_flush_exception_returns_false(self) -> None:
        mock_exp = MagicMock()
        mock_exp.force_flush.side_effect = RuntimeError("flush failed")
        exporter = AgentSpanExporter(exporters=[mock_exp])
        assert exporter.force_flush() is False

    def test_agent_id_session_id_stored(self) -> None:
        exporter = AgentSpanExporter(agent_id="agent-1", session_id="session-1")
        assert exporter._agent_id == "agent-1"
        assert exporter._session_id == "session-1"


# ── build_otlp_exporter ────────────────────────────────────────────────────────


class TestBuildOtlpExporter:
    def test_returns_console_when_no_otlp_available(self) -> None:
        with patch("agent_observability.tracing.exporter._OTLP_GRPC_AVAILABLE", False):
            with patch("agent_observability.tracing.exporter._OTLP_HTTP_AVAILABLE", False):
                result = build_otlp_exporter()
        assert isinstance(result, ConsoleSpanExporter)

    def test_returns_http_when_available(self) -> None:
        mock_http_cls = MagicMock()
        mock_http_instance = MagicMock()
        mock_http_cls.return_value = mock_http_instance

        with patch("agent_observability.tracing.exporter._OTLP_HTTP_AVAILABLE", True):
            with patch("agent_observability.tracing.exporter._HttpOTLP", mock_http_cls):
                result = build_otlp_exporter(endpoint="http://localhost:4318", use_http=True)
        assert result is mock_http_instance

    def test_returns_grpc_when_available(self) -> None:
        mock_grpc_cls = MagicMock()
        mock_grpc_instance = MagicMock()
        mock_grpc_cls.return_value = mock_grpc_instance

        with patch("agent_observability.tracing.exporter._OTLP_GRPC_AVAILABLE", True):
            with patch("agent_observability.tracing.exporter._GrpcOTLP", mock_grpc_cls):
                result = build_otlp_exporter(endpoint="http://localhost:4317", use_http=False)
        assert result is mock_grpc_instance


# ── _NoOpTracer / _NoOpSpanHandle ──────────────────────────────────────────────


class TestNoOpTracer:
    def test_start_span_returns_handle(self) -> None:
        tracer = _NoOpTracer()
        handle = tracer.start_span("test-span")
        assert isinstance(handle, _NoOpSpanHandle)
        assert handle.name == "test-span"

    def test_start_as_current_span_is_context_manager(self) -> None:
        tracer = _NoOpTracer()
        ctx = tracer.start_as_current_span("test-span")
        with ctx as span:
            assert isinstance(span, _NoOpSpanHandle)


class TestNoOpSpanHandle:
    def test_set_attribute_noop(self) -> None:
        handle = _NoOpSpanHandle("test")
        handle.set_attribute("key", "value")  # should not raise

    def test_record_exception_noop(self) -> None:
        handle = _NoOpSpanHandle("test")
        handle.record_exception(RuntimeError("err"))  # should not raise

    def test_set_status_noop(self) -> None:
        handle = _NoOpSpanHandle("test")
        handle.set_status("OK")  # should not raise

    def test_end_noop(self) -> None:
        handle = _NoOpSpanHandle("test")
        handle.end()  # should not raise

    def test_context_manager(self) -> None:
        handle = _NoOpSpanHandle("test")
        with handle as span:
            assert span is handle


# ── AgentTracerProvider ────────────────────────────────────────────────────────


class TestAgentTracerProvider:
    def test_build_no_otel_returns_none(self) -> None:
        provider = AgentTracerProvider(service_name="test-svc")
        with patch("agent_observability.tracing.tracer._OTEL_SDK_AVAILABLE", False):
            result = provider.build()
        assert result is None

    def test_get_tracer_no_otel_returns_noop(self) -> None:
        provider = AgentTracerProvider()
        with patch("agent_observability.tracing.tracer._OTEL_SDK_AVAILABLE", False):
            tracer = provider.get_tracer()
        assert isinstance(tracer, _NoOpTracer)

    def test_get_tracer_with_otel(self) -> None:
        provider = AgentTracerProvider()
        mock_trace = MagicMock()
        mock_tracer = MagicMock()
        mock_trace.get_tracer.return_value = mock_tracer
        with patch("agent_observability.tracing.tracer._OTEL_SDK_AVAILABLE", True):
            with patch("agent_observability.tracing.tracer.otel_trace", mock_trace):
                tracer = provider.get_tracer("my-lib")
        assert tracer is mock_tracer

    def test_shutdown_no_provider(self) -> None:
        provider = AgentTracerProvider()
        provider.shutdown()  # should not raise when _provider is None

    def test_shutdown_calls_provider_shutdown(self) -> None:
        provider = AgentTracerProvider()
        mock_otel_provider = MagicMock()
        provider._provider = mock_otel_provider
        provider.shutdown()
        mock_otel_provider.shutdown.assert_called_once()

    def test_build_with_otel_sdk(self) -> None:
        mock_resource = MagicMock()
        mock_resource.create.return_value = MagicMock()
        mock_provider_cls = MagicMock()
        mock_provider_instance = MagicMock()
        mock_provider_cls.return_value = mock_provider_instance
        mock_batch_processor = MagicMock()
        mock_simple_processor = MagicMock()
        mock_otel_trace = MagicMock()

        with patch("agent_observability.tracing.tracer._OTEL_SDK_AVAILABLE", True):
            with patch("agent_observability.tracing.tracer.Resource", mock_resource):
                with patch("agent_observability.tracing.tracer.TracerProvider", mock_provider_cls):
                    with patch("agent_observability.tracing.tracer.BatchSpanProcessor", mock_batch_processor):
                        with patch("agent_observability.tracing.tracer.SimpleSpanProcessor", mock_simple_processor):
                            with patch("agent_observability.tracing.tracer.otel_trace", mock_otel_trace):
                                provider = AgentTracerProvider(
                                    service_name="test-svc",
                                    agent_id="agent-1",
                                )
                                mock_exporter = MagicMock()
                                result = provider.build(exporter=mock_exporter, batch=True)
        assert result is mock_provider_instance


# ── setup_tracing ──────────────────────────────────────────────────────────────


class TestSetupTracing:
    def test_returns_agent_tracer_provider(self) -> None:
        with patch("agent_observability.tracing.tracer._OTEL_SDK_AVAILABLE", False):
            result = setup_tracing(service_name="test")
        assert isinstance(result, AgentTracerProvider)

    def test_console_exporter_when_no_downstream(self) -> None:
        with patch("agent_observability.tracing.tracer._OTEL_SDK_AVAILABLE", False):
            provider = setup_tracing(service_name="test")
        # With no OTel SDK, build returns None but provider is created
        assert provider._service_name == "test"

    def test_jsonl_path_creates_jsonl_exporter(self) -> None:
        from agent_observability.tracing.exporter import JsonLinesExporter

        created_exporters: list[Any] = []

        original_init = AgentSpanExporter.__init__

        def capturing_init(self: AgentSpanExporter, exporters: Any = None, **kwargs: Any) -> None:
            if exporters:
                created_exporters.extend(exporters)
            original_init(self, exporters=exporters, **kwargs)

        with patch("agent_observability.tracing.tracer._OTEL_SDK_AVAILABLE", False):
            with patch.object(AgentSpanExporter, "__init__", capturing_init):
                setup_tracing(service_name="test", jsonl_path="/tmp/spans.jsonl")

        jsonl_exporters = [e for e in created_exporters if isinstance(e, JsonLinesExporter)]
        assert len(jsonl_exporters) == 1

    def test_console_flag_adds_console_exporter(self) -> None:
        created_exporters: list[Any] = []

        original_init = AgentSpanExporter.__init__

        def capturing_init(self: AgentSpanExporter, exporters: Any = None, **kwargs: Any) -> None:
            if exporters:
                created_exporters.extend(exporters)
            original_init(self, exporters=exporters, **kwargs)

        with patch("agent_observability.tracing.tracer._OTEL_SDK_AVAILABLE", False):
            with patch.object(AgentSpanExporter, "__init__", capturing_init):
                setup_tracing(service_name="test", console=True)

        console_exporters = [e for e in created_exporters if isinstance(e, ConsoleSpanExporter)]
        assert len(console_exporters) >= 1



"""Tests for CorrelationContext."""
from __future__ import annotations

import pytest

from agent_observability.correlation.correlation_context import (
    BaggageItem,
    CorrelationContext,
)


class TestBaggageItem:
    def test_basic_construction(self) -> None:
        item = BaggageItem(key="tenant", value="acme-corp")
        assert item.key == "tenant"
        assert item.value == "acme-corp"

    def test_empty_key_raises(self) -> None:
        with pytest.raises(ValueError, match="key"):
            BaggageItem(key="", value="value")

    def test_is_frozen(self) -> None:
        item = BaggageItem(key="tenant", value="acme")
        with pytest.raises(Exception):
            item.key = "other"  # type: ignore[misc]

    def test_to_dict(self) -> None:
        item = BaggageItem(key="user", value="alice")
        d = item.to_dict()
        assert d == {"key": "user", "value": "alice"}


class TestCorrelationContext:
    def test_new_root_generates_ids(self) -> None:
        ctx = CorrelationContext.new_root()
        assert ctx.trace_id != ""
        assert ctx.span_id != ""
        assert ctx.parent_span_id is None

    def test_new_root_is_root(self) -> None:
        ctx = CorrelationContext.new_root()
        assert ctx.is_root is True

    def test_new_root_with_custom_trace_id(self) -> None:
        ctx = CorrelationContext.new_root(trace_id="custom-trace")
        assert ctx.trace_id == "custom-trace"

    def test_new_child_span_inherits_trace_id(self) -> None:
        parent = CorrelationContext.new_root()
        child = parent.new_child_span()
        assert child.trace_id == parent.trace_id

    def test_new_child_span_sets_parent(self) -> None:
        parent = CorrelationContext.new_root()
        child = parent.new_child_span()
        assert child.parent_span_id == parent.span_id

    def test_new_child_span_is_not_root(self) -> None:
        parent = CorrelationContext.new_root()
        child = parent.new_child_span()
        assert child.is_root is False

    def test_new_child_span_custom_span_id(self) -> None:
        parent = CorrelationContext.new_root()
        child = parent.new_child_span(span_id="custom-child")
        assert child.span_id == "custom-child"

    def test_different_roots_have_different_trace_ids(self) -> None:
        ctx1 = CorrelationContext.new_root()
        ctx2 = CorrelationContext.new_root()
        assert ctx1.trace_id != ctx2.trace_id

    def test_get_baggage_returns_value(self) -> None:
        ctx = CorrelationContext.new_root()
        ctx_with_baggage = ctx.with_baggage("tenant", "acme")
        assert ctx_with_baggage.get_baggage("tenant") == "acme"

    def test_get_baggage_returns_none_for_missing(self) -> None:
        ctx = CorrelationContext.new_root()
        assert ctx.get_baggage("nonexistent") is None

    def test_with_baggage_returns_new_context(self) -> None:
        ctx = CorrelationContext.new_root()
        ctx2 = ctx.with_baggage("key", "value")
        assert ctx2 is not ctx
        assert ctx.get_baggage("key") is None
        assert ctx2.get_baggage("key") == "value"

    def test_with_baggage_replaces_existing_key(self) -> None:
        ctx = CorrelationContext.new_root()
        ctx = ctx.with_baggage("key", "old")
        ctx = ctx.with_baggage("key", "new")
        assert ctx.get_baggage("key") == "new"
        assert len(ctx.baggage) == 1

    def test_baggage_propagated_to_child(self) -> None:
        ctx = CorrelationContext.new_root()
        ctx = ctx.with_baggage("tenant", "corp")
        child = ctx.new_child_span()
        assert child.get_baggage("tenant") == "corp"

    def test_to_dict_structure(self) -> None:
        ctx = CorrelationContext.new_root()
        d = ctx.to_dict()
        assert "trace_id" in d
        assert "span_id" in d
        assert "parent_span_id" in d
        assert "sampled" in d
        assert "baggage" in d

    def test_from_dict_round_trip(self) -> None:
        ctx = CorrelationContext.new_root()
        ctx = ctx.with_baggage("tenant", "acme")
        d = ctx.to_dict()
        reconstructed = CorrelationContext.from_dict(d)
        assert reconstructed.trace_id == ctx.trace_id
        assert reconstructed.span_id == ctx.span_id
        assert reconstructed.get_baggage("tenant") == "acme"

    def test_from_dict_raises_on_missing_trace_id(self) -> None:
        with pytest.raises(ValueError):
            CorrelationContext.from_dict({"span_id": "s1"})

    def test_from_dict_raises_on_missing_span_id(self) -> None:
        with pytest.raises(ValueError):
            CorrelationContext.from_dict({"trace_id": "t1"})

    def test_to_headers_produces_traceparent(self) -> None:
        ctx = CorrelationContext.new_root()
        headers = ctx.to_headers()
        assert "traceparent" in headers
        assert headers["traceparent"].startswith("00-")

    def test_to_headers_includes_baggage(self) -> None:
        ctx = CorrelationContext.new_root()
        ctx = ctx.with_baggage("tenant", "acme")
        headers = ctx.to_headers()
        assert "baggage" in headers
        assert "tenant=acme" in headers["baggage"]

    def test_sampled_false_propagated_to_child(self) -> None:
        ctx = CorrelationContext.new_root(sampled=False)
        child = ctx.new_child_span()
        assert child.sampled is False



"""PiiDetector — regex-based PII detection in span attributes and strings."""
from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Pattern, Sequence

from agent_observability.privacy.patterns import (
    PiiPattern,
    get_patterns_for_jurisdiction,
)


@dataclass
class PiiMatch:
    """A single PII finding within a text value."""

    pattern_name: str
    replacement_token: str
    start: int
    end: int
    matched_text: str


@dataclass
class DetectionResult:
    """Result of scanning a single text value for PII."""

    original: str
    matches: list[PiiMatch]

    @property
    def has_pii(self) -> bool:
        return bool(self.matches)

    @property
    def pattern_names(self) -> list[str]:
        return [m.pattern_name for m in self.matches]


class PiiDetector:
    """Scan text values for PII using compiled regex patterns.

    Parameters
    ----------
    jurisdiction:
        Which predefined pattern set to use (``"common"``, ``"us"``,
        ``"eu"``, ``"all"``).
    extra_patterns:
        Additional :class:`~agent_observability.privacy.patterns.PiiPattern`
        objects to include beyond the jurisdiction defaults.
    """

    def __init__(
        self,
        jurisdiction: str = "all",
        extra_patterns: Sequence[PiiPattern] | None = None,
    ) -> None:
        base = get_patterns_for_jurisdiction(jurisdiction)
        all_patterns = list(base) + list(extra_patterns or [])
        # De-duplicate by name (last writer wins)
        seen: dict[str, PiiPattern] = {}
        for p in all_patterns:
            seen[p.name] = p
        self._patterns: list[tuple[PiiPattern, Pattern[str]]] = [
            (p, p.compiled()) for p in seen.values()
        ]

    # ── Detection ─────────────────────────────────────────────────────────────

    def detect(self, text: str) -> DetectionResult:
        """Scan *text* for PII and return all matches.

        Parameters
        ----------
        text:
            The string to scan.

        Returns
        -------
        A :class:`DetectionResult` containing all :class:`PiiMatch` objects.
        """
        matches: list[PiiMatch] = []
        for pii_pattern, compiled in self._patterns:
            for m in compiled.finditer(text):
                matches.append(
                    PiiMatch(
                        pattern_name=pii_pattern.name,
                        replacement_token=pii_pattern.replacement_token,
                        start=m.start(),
                        end=m.end(),
                        matched_text=m.group(0),
                    )
                )
        # Sort by position for deterministic output
        matches.sort(key=lambda x: x.start)
        return DetectionResult(original=text, matches=matches)

    def has_pii(self, text: str) -> bool:
        """Return ``True`` if *text* contains any recognised PII."""
        for _, compiled in self._patterns:
            if compiled.search(text):
                return True
        return False

    def scan_dict(
        self,
        data: dict[str, object],
        exclude_keys: Sequence[str] | None = None,
    ) -> dict[str, DetectionResult]:
        """Scan all string values in a flat dict.

        Parameters
        ----------
        data:
            A flat ``{key: value}`` dict (e.g. span attributes).
        exclude_keys:
            Keys to skip during scanning.

        Returns
        -------
        ``{key: DetectionResult}`` for every string value that was scanned.
        """
        excluded = set(exclude_keys or [])
        results: dict[str, DetectionResult] = {}
        for key, value in data.items():
            if key in excluded or not isinstance(value, str):
                continue
            result = self.detect(value)
            if result.has_pii:
                results[key] = result
        return results



"""AgentTracerProvider — configure OTel with agent-specific span processors.

``setup_tracing()`` is the single entry point that most users will call.
It wires up the provider, sampler, enricher, and exporter in one shot.
"""
from __future__ import annotations

import logging
from typing import Optional

from agent_observability.spans.enricher import SpanEnricher
from agent_observability.tracing.exporter import (
    AgentSpanExporter,
    ConsoleSpanExporter,
    JsonLinesExporter,
    build_otlp_exporter,
)
from agent_observability.tracing.sampler import CostAwareSampler

logger = logging.getLogger(__name__)

try:
    from opentelemetry import trace as otel_trace
    from opentelemetry.sdk.resources import Resource
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.sdk.trace.export import BatchSpanProcessor, SimpleSpanProcessor

    _OTEL_SDK_AVAILABLE = True
except ImportError:
    _OTEL_SDK_AVAILABLE = False
    otel_trace = None  # type: ignore[assignment]
    Resource = None  # type: ignore[assignment,misc]
    TracerProvider = None  # type: ignore[assignment,misc]
    BatchSpanProcessor = None  # type: ignore[assignment,misc]
    SimpleSpanProcessor = None  # type: ignore[assignment,misc]


class AgentTracerProvider:
    """Thin wrapper around the OTel ``TracerProvider`` adding agent conveniences.

    Parameters
    ----------
    service_name:
        Reported in OTel resource attributes.
    service_version:
        Reported in OTel resource attributes.
    agent_id:
        Injected into every span via :class:`SpanEnricher`.
    session_id:
        Injected into every span via :class:`SpanEnricher`.
    framework:
        Injected into every span (e.g. ``"langchain"``).
    environment:
        Deployment environment (``"production"``, ``"staging"``, …).
    """

    def __init__(
        self,
        service_name: str = "agent",
        service_version: str = "0.0.0",
        agent_id: str = "",
        session_id: str = "",
        framework: str = "",
        environment: str = "",
    ) -> None:
        self._service_name = service_name
        self._service_version = service_version
        self._agent_id = agent_id
        self._session_id = session_id
        self._framework = framework
        self._environment = environment
        self._provider: Optional[object] = None
        self._enricher: Optional[SpanEnricher] = None

    def build(
        self,
        exporter: Optional[object] = None,
        sampler: Optional[object] = None,
        batch: bool = True,
    ) -> Optional[object]:
        """Create and register the OTel TracerProvider.

        Parameters
        ----------
        exporter:
            A SpanExporter instance.  Defaults to :class:`ConsoleSpanExporter`.
        sampler:
            A Sampler instance.  Defaults to :class:`CostAwareSampler`.
        batch:
            Use ``BatchSpanProcessor`` when ``True``, ``SimpleSpanProcessor``
            when ``False``.

        Returns
        -------
        The configured ``TracerProvider``, or ``None`` if OTel SDK is absent.
        """
        if not _OTEL_SDK_AVAILABLE:
            logger.warning(
                "opentelemetry-sdk not installed; AgentTracerProvider.build() is a no-op"
            )
            return None

        assert Resource is not None
        assert TracerProvider is not None
        assert BatchSpanProcessor is not None
        assert SimpleSpanProcessor is not None

        resource = Resource.create(
            {
                "service.name": self._service_name,
                "service.version": self._service_version,
                "agent.id": self._agent_id,
                "deployment.environment": self._environment,
            }
        )

        chosen_sampler = sampler or CostAwareSampler()
        provider = TracerProvider(resource=resource, sampler=chosen_sampler)

        # Span enricher processor
        self._enricher = SpanEnricher(
            agent_id=self._agent_id,
            session_id=self._session_id,
            framework=self._framework,
            environment=self._environment,
        )
        provider.add_span_processor(self._enricher)

        # Export processor
        chosen_exporter = exporter or ConsoleSpanExporter()
        processor_cls = BatchSpanProcessor if batch else SimpleSpanProcessor
        provider.add_span_processor(processor_cls(chosen_exporter))

        otel_trace.set_tracer_provider(provider)  # type: ignore[union-attr]
        self._provider = provider
        logger.info(
            "AgentTracerProvider configured for service '%s'", self._service_name
        )
        return provider

    def get_tracer(self, name: str = "agent-observability") -> object:
        """Return a tracer scoped to *name*."""
        if not _OTEL_SDK_AVAILABLE or otel_trace is None:
            return _NoOpTracer()
        return otel_trace.get_tracer(name)

    def shutdown(self) -> None:
        """Flush and shut down all span processors."""
        if self._provider is not None and hasattr(self._provider, "shutdown"):
            self._provider.shutdown()  # type: ignore[union-attr]


class _NoOpTracer:
    """Fallback tracer used when OTel SDK is not installed."""

    def start_span(self, name: str, **kwargs: object) -> object:
        return _NoOpSpanHandle(name)

    def start_as_current_span(self, name: str, **kwargs: object) -> object:
        from contextlib import contextmanager

        @contextmanager  # type: ignore[misc]
        def _ctx() -> object:  # type: ignore[misc]
            yield _NoOpSpanHandle(name)

        return _ctx()


class _NoOpSpanHandle:
    def __init__(self, name: str) -> None:
        self.name = name

    def set_attribute(self, key: str, value: object) -> None:
        pass

    def record_exception(self, exc: BaseException) -> None:
        pass

    def set_status(self, status: object, description: str = "") -> None:
        pass

    def end(self) -> None:
        pass

    def __enter__(self) -> "_NoOpSpanHandle":
        return self

    def __exit__(self, *args: object) -> None:
        pass


def setup_tracing(
    service_name: str = "agent",
    service_version: str = "0.0.0",
    agent_id: str = "",
    session_id: str = "",
    framework: str = "",
    environment: str = "development",
    otlp_endpoint: str = "",
    otlp_use_http: bool = False,
    jsonl_path: str = "",
    console: bool = False,
    batch: bool = True,
    high_cost_threshold_usd: float = 0.01,
    low_cost_sample_rate: float = 0.1,
) -> AgentTracerProvider:
    """One-shot helper to configure agent observability tracing.

    Parameters
    ----------
    service_name:
        OTel resource ``service.name``.
    service_version:
        OTel resource ``service.version``.
    agent_id:
        Agent identifier stamped on every span.
    session_id:
        Session identifier stamped on every span.
    framework:
        Framework name stamped on every span.
    environment:
        Deployment environment.
    otlp_endpoint:
        If non-empty, send spans to this OTLP collector endpoint.
    otlp_use_http:
        Use HTTP transport for OTLP (default: gRPC).
    jsonl_path:
        If non-empty, also write spans to this JSON Lines file.
    console:
        If ``True``, also emit spans to the console logger.
    batch:
        Use ``BatchSpanProcessor`` (recommended for production).
    high_cost_threshold_usd:
        Traces above this cost are always sampled.
    low_cost_sample_rate:
        Fraction of cheap traces to keep.

    Returns
    -------
    The configured :class:`AgentTracerProvider`.
    """
    provider = AgentTracerProvider(
        service_name=service_name,
        service_version=service_version,
        agent_id=agent_id,
        session_id=session_id,
        framework=framework,
        environment=environment,
    )

    # Build the fan-out exporter
    downstream_exporters: list[object] = []
    if otlp_endpoint:
        downstream_exporters.append(
            build_otlp_exporter(
                endpoint=otlp_endpoint,
                use_http=otlp_use_http,
            )
        )
    if jsonl_path:
        downstream_exporters.append(JsonLinesExporter(path=jsonl_path))
    if console or not downstream_exporters:
        downstream_exporters.append(ConsoleSpanExporter())

    fan_out = AgentSpanExporter(
        exporters=downstream_exporters,  # type: ignore[arg-type]
        agent_id=agent_id,
        session_id=session_id,
    )

    sampler = CostAwareSampler(
        high_cost_threshold_usd=high_cost_threshold_usd,
        low_cost_sample_rate=low_cost_sample_rate,
    )

    provider.build(exporter=fan_out, sampler=sampler, batch=batch)
    return provider



"""Tests for pii.redactor (the legacy standalone PII redactor module)."""
from __future__ import annotations

import pytest

from agent_observability.pii.redactor import PiiRedactor, RedactionConfig


class TestRedactionConfig:
    def test_defaults(self) -> None:
        config = RedactionConfig()
        assert config.redact_emails is True
        assert config.redact_phone is True
        assert config.redact_ssn is True
        assert config.redact_credit_card is True
        assert config.redact_ip is True
        assert config.redact_custom_patterns == []
        assert config.replacement == "[REDACTED]"

    def test_custom_replacement(self) -> None:
        config = RedactionConfig(replacement="***")
        assert config.replacement == "***"


class TestPiiRedactorLegacy:
    def test_redact_email(self) -> None:
        redactor = PiiRedactor()
        result = redactor.redact("contact user@example.com today")
        assert "user@example.com" not in result
        assert "[REDACTED]" in result

    def test_redact_ssn(self) -> None:
        redactor = PiiRedactor()
        result = redactor.redact("ssn: 123-45-6789")
        assert "123-45-6789" not in result

    def test_redact_credit_card_visa(self) -> None:
        redactor = PiiRedactor()
        result = redactor.redact("card 4111111111111111 charged")
        assert "4111111111111111" not in result

    def test_redact_ipv4(self) -> None:
        redactor = PiiRedactor()
        result = redactor.redact("server 192.168.0.1 pinged")
        assert "192.168.0.1" not in result

    def test_no_redaction_for_clean_text(self) -> None:
        redactor = PiiRedactor()
        text = "nothing sensitive at all"
        assert redactor.redact(text) == text

    def test_redact_disabled_fields(self) -> None:
        config = RedactionConfig(
            redact_emails=False,
            redact_phone=False,
            redact_ssn=False,
            redact_credit_card=False,
            redact_ip=False,
        )
        redactor = PiiRedactor(config=config)
        text = "user@example.com 123-45-6789 192.168.1.1"
        # No patterns active — text should pass through
        assert redactor.redact(text) == text

    def test_redact_span_attributes(self) -> None:
        redactor = PiiRedactor()
        attributes = {
            "user_email": "user@example.com",
            "count": 42,
            "safe": "nothing here",
        }
        result = redactor.redact_span_attributes(attributes)
        assert "user@example.com" not in result["user_email"]
        assert result["count"] == 42
        assert result["safe"] == "nothing here"

    def test_redact_span_attributes_non_string_pass_through(self) -> None:
        redactor = PiiRedactor()
        attributes = {"number": 123, "flag": True, "items": [1, 2, 3]}
        result = redactor.redact_span_attributes(attributes)
        assert result["number"] == 123
        assert result["flag"] is True

    def test_redact_dict_shallow(self) -> None:
        redactor = PiiRedactor()
        data = {"email": "user@example.com"}
        result = redactor.redact_dict(data)
        assert "user@example.com" not in result["email"]

    def test_redact_dict_nested(self) -> None:
        redactor = PiiRedactor()
        data = {"outer": {"inner": "user@example.com"}}
        result = redactor.redact_dict(data)
        inner = result["outer"]
        assert isinstance(inner, dict)
        assert "user@example.com" not in inner["inner"]

    def test_redact_dict_with_list(self) -> None:
        redactor = PiiRedactor()
        data = {"emails": ["user@a.com", "user@b.com"]}
        result = redactor.redact_dict(data)
        for email in result["emails"]:  # type: ignore[union-attr]
            assert "user@" not in email

    def test_redact_dict_non_string_leaf_passthrough(self) -> None:
        redactor = PiiRedactor()
        data = {"value": 42}
        result = redactor.redact_dict(data)
        assert result["value"] == 42

    def test_custom_patterns_applied(self) -> None:
        config = RedactionConfig(redact_custom_patterns=[r"SECRET-\d+"])
        redactor = PiiRedactor(config=config)
        result = redactor.redact("token SECRET-9876 used")
        assert "SECRET-9876" not in result

    def test_invalid_custom_pattern_skipped(self) -> None:
        config = RedactionConfig(redact_custom_patterns=[r"[invalid("])
        # Should not raise during init
        redactor = PiiRedactor(config=config)
        result = redactor.redact("some text")
        assert isinstance(result, str)

    def test_custom_replacement_token(self) -> None:
        config = RedactionConfig(replacement="***")
        redactor = PiiRedactor(config=config)
        result = redactor.redact("email user@example.com here")
        assert "***" in result



"""AgentMetricCollector — collect and store agent metrics in memory.

Provides a thread-safe, dependency-free metrics accumulator.  Can feed
:class:`~agent_observability.metrics.exporter.PrometheusExporter`.
"""
from __future__ import annotations

import math
import threading
import time
from dataclasses import dataclass, field
from typing import Optional, Sequence


@dataclass
class CounterValue:
    """Monotonically increasing counter with labels."""

    name: str
    labels: dict[str, str]
    value: float = 0.0

    def inc(self, amount: float = 1.0) -> None:
        self.value += amount


@dataclass
class GaugeValue:
    """Gauge (can go up or down) with labels."""

    name: str
    labels: dict[str, str]
    value: float = 0.0

    def set(self, value: float) -> None:
        self.value = value

    def inc(self, amount: float = 1.0) -> None:
        self.value += amount

    def dec(self, amount: float = 1.0) -> None:
        self.value -= amount


@dataclass
class HistogramValue:
    """Simplified histogram with pre-defined buckets."""

    name: str
    labels: dict[str, str]
    # Bucket upper bounds
    buckets: tuple[float, ...] = (0.005, 0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1.0, 2.5, 5.0, 10.0)
    _counts: list[int] = field(default_factory=list)
    _sum: float = 0.0
    _total_count: int = 0

    def __post_init__(self) -> None:
        self._counts = [0] * len(self.buckets)

    def observe(self, value: float) -> None:
        self._sum += value
        self._total_count += 1
        for i, upper_bound in enumerate(self.buckets):
            if value <= upper_bound:
                self._counts[i] += 1

    @property
    def count(self) -> int:
        return self._total_count

    @property
    def sum_value(self) -> float:
        return self._sum

    def bucket_counts(self) -> list[tuple[float, int]]:
        """Return list of ``(upper_bound, cumulative_count)`` pairs."""
        cumulative = 0
        result: list[tuple[float, int]] = []
        for upper_bound, count in zip(self.buckets, self._counts):
            cumulative += count
            result.append((upper_bound, cumulative))
        return result


_LabelKey = tuple[str, ...]


class AgentMetricCollector:
    """Thread-safe in-memory metric store.

    All metric types are stored as dicts keyed by ``(metric_name, label_tuple)``.
    """

    def __init__(self) -> None:
        self._lock = threading.Lock()
        self._counters: dict[tuple[str, _LabelKey], CounterValue] = {}
        self._gauges: dict[tuple[str, _LabelKey], GaugeValue] = {}
        self._histograms: dict[tuple[str, _LabelKey], HistogramValue] = {}

    # ── Counters ──────────────────────────────────────────────────────────────

    def increment_counter(
        self,
        name: str,
        labels: Optional[dict[str, str]] = None,
        amount: float = 1.0,
    ) -> None:
        """Increment a named counter."""
        label_dict = labels or {}
        key = (name, tuple(sorted(label_dict.items())))
        with self._lock:
            if key not in self._counters:
                self._counters[key] = CounterValue(name=name, labels=label_dict)
            self._counters[key].inc(amount)

    # ── Gauges ────────────────────────────────────────────────────────────────

    def set_gauge(
        self,
        name: str,
        value: float,
        labels: Optional[dict[str, str]] = None,
    ) -> None:
        """Set a gauge to an absolute value."""
        label_dict = labels or {}
        key = (name, tuple(sorted(label_dict.items())))
        with self._lock:
            if key not in self._gauges:
                self._gauges[key] = GaugeValue(name=name, labels=label_dict)
            self._gauges[key].set(value)

    def increment_gauge(
        self,
        name: str,
        labels: Optional[dict[str, str]] = None,
        amount: float = 1.0,
    ) -> None:
        label_dict = labels or {}
        key = (name, tuple(sorted(label_dict.items())))
        with self._lock:
            if key not in self._gauges:
                self._gauges[key] = GaugeValue(name=name, labels=label_dict)
            self._gauges[key].inc(amount)

    def decrement_gauge(
        self,
        name: str,
        labels: Optional[dict[str, str]] = None,
        amount: float = 1.0,
    ) -> None:
        label_dict = labels or {}
        key = (name, tuple(sorted(label_dict.items())))
        with self._lock:
            if key not in self._gauges:
                self._gauges[key] = GaugeValue(name=name, labels=label_dict)
            self._gauges[key].dec(amount)

    # ── Histograms ────────────────────────────────────────────────────────────

    def observe_histogram(
        self,
        name: str,
        value: float,
        labels: Optional[dict[str, str]] = None,
    ) -> None:
        """Record a histogram observation."""
        label_dict = labels or {}
        key = (name, tuple(sorted(label_dict.items())))
        with self._lock:
            if key not in self._histograms:
                self._histograms[key] = HistogramValue(name=name, labels=label_dict)
            self._histograms[key].observe(value)

    # ── Domain helpers ────────────────────────────────────────────────────────

    def record_llm_call(
        self,
        agent_id: str,
        model: str,
        provider: str,
        status: str,
        latency_seconds: float,
        input_tokens: int,
        output_tokens: int,
        cost_usd: float,
    ) -> None:
        """Record all metrics associated with a single LLM call."""
        labels = {"agent_id": agent_id, "model": model, "provider": provider, "status": status}
        self.increment_counter("agent_llm_calls_total", labels)
        self.observe_histogram(
            "agent_latency_seconds",
            latency_seconds,
            {"agent_id": agent_id, "operation": "llm_call"},
        )
        self.increment_counter(
            "agent_tokens_total",
            {"agent_id": agent_id, "model": model, "token_type": "input"},
            amount=float(input_tokens),
        )
        self.increment_counter(
            "agent_tokens_total",
            {"agent_id": agent_id, "model": model, "token_type": "output"},
            amount=float(output_tokens),
        )
        cost_labels = {"agent_id": agent_id, "model": model, "provider": provider}
        self.increment_counter("agent_cost_usd_total", cost_labels, amount=cost_usd)

    def record_tool_invocation(
        self,
        agent_id: str,
        tool_name: str,
        success: bool,
        latency_seconds: float,
    ) -> None:
        """Record metrics for a tool invocation."""
        labels = {"agent_id": agent_id, "tool_name": tool_name, "success": str(success).lower()}
        self.increment_counter("agent_tool_invocations_total", labels)
        self.observe_histogram(
            "agent_latency_seconds",
            latency_seconds,
            {"agent_id": agent_id, "operation": "tool_invoke"},
        )

    def record_error(
        self,
        agent_id: str,
        error_type: str,
        recoverable: bool,
    ) -> None:
        """Record an agent error."""
        labels = {
            "agent_id": agent_id,
            "error_type": error_type,
            "recoverable": str(recoverable).lower(),
        }
        self.increment_counter("agent_errors_total", labels)

    def record_memory_operation(
        self,
        agent_id: str,
        operation: str,
        backend: str,
    ) -> None:
        """Record a memory read or write operation."""
        self.increment_counter(
            "agent_memory_operations_total",
            {"agent_id": agent_id, "operation": operation, "backend": backend},
        )

    def set_drift_z_score(self, agent_id: str, z_score: float) -> None:
        """Update the drift Z-score gauge for an agent."""
        self.set_gauge("agent_drift_z_score", z_score, {"agent_id": agent_id})

    # ── Snapshot ──────────────────────────────────────────────────────────────

    def snapshot(self) -> dict[str, object]:
        """Return a serialisable snapshot of all current metric values."""
        with self._lock:
            return {
                "counters": [
                    {"name": v.name, "labels": v.labels, "value": v.value}
                    for v in self._counters.values()
                ],
                "gauges": [
                    {"name": v.name, "labels": v.labels, "value": v.value}
                    for v in self._gauges.values()
                ],
                "histograms": [
                    {
                        "name": v.name,
                        "labels": v.labels,
                        "count": v.count,
                        "sum": v.sum_value,
                        "buckets": v.bucket_counts(),
                    }
                    for v in self._histograms.values()
                ],
            }

    def reset(self) -> None:
        """Clear all metrics (primarily for testing)."""
        with self._lock:
            self._counters.clear()
            self._gauges.clear()
            self._histograms.clear()



"""DriftAlertManager — fire alerts when agent drift is detected.

Supports three output channels:
- Console (Rich/logging)
- Application log (Python ``logging``)
- Webhook (HTTP POST)
"""
from __future__ import annotations

import json
import logging
import time
import urllib.request
from dataclasses import asdict, dataclass
from typing import Callable, Optional, Sequence

from agent_observability.drift.detector import DriftResult

logger = logging.getLogger(__name__)


AlertHandler = Callable[["DriftAlert"], None]


@dataclass
class DriftAlert:
    """An alert fired when drift is detected."""

    agent_id: str
    severity: str  # "low", "medium", "high"
    max_z_score: float
    threshold: float
    drifted_features: dict[str, float]
    baseline_age_seconds: float
    window_span_count: int
    timestamp: float
    message: str

    def to_dict(self) -> dict[str, object]:
        return {
            "agent_id": self.agent_id,
            "severity": self.severity,
            "max_z_score": self.max_z_score,
            "threshold": self.threshold,
            "drifted_features": self.drifted_features,
            "baseline_age_seconds": self.baseline_age_seconds,
            "window_span_count": self.window_span_count,
            "timestamp": self.timestamp,
            "message": self.message,
        }


def _result_to_alert(result: DriftResult) -> DriftAlert:
    top_features = sorted(
        result.drifted_features.items(), key=lambda x: abs(x[1]), reverse=True
    )[:3]
    feature_summary = ", ".join(
        f"{k}={v:.2f}σ" for k, v in top_features
    )
    message = (
        f"Agent '{result.agent_id}' drifted ({result.severity.upper()}): "
        f"max_z={result.max_z_score:.2f}σ. "
        f"Top features: {feature_summary or 'none'}."
    )
    return DriftAlert(
        agent_id=result.agent_id,
        severity=result.severity,
        max_z_score=result.max_z_score,
        threshold=result.threshold,
        drifted_features=result.drifted_features,
        baseline_age_seconds=result.baseline_age_seconds,
        window_span_count=result.window_span_count,
        timestamp=result.timestamp,
        message=message,
    )


class ConsoleAlertHandler:
    """Write alerts to the Python logger (or Rich console)."""

    def __call__(self, alert: DriftAlert) -> None:
        log_fn = {
            "low": logger.info,
            "medium": logger.warning,
            "high": logger.error,
        }.get(alert.severity, logger.warning)
        log_fn("[DRIFT ALERT] %s", alert.message)


class WebhookAlertHandler:
    """HTTP POST the alert payload to a webhook URL.

    Parameters
    ----------
    url:
        The webhook endpoint to POST to.
    timeout_seconds:
        Request timeout in seconds.
    extra_headers:
        Additional HTTP headers to include (e.g. authorization).
    """

    def __init__(
        self,
        url: str,
        timeout_seconds: float = 5.0,
        extra_headers: Optional[dict[str, str]] = None,
    ) -> None:
        self._url = url
        self._timeout = timeout_seconds
        self._extra_headers = extra_headers or {}

    def __call__(self, alert: DriftAlert) -> None:
        payload = json.dumps(alert.to_dict()).encode("utf-8")
        headers = {"Content-Type": "application/json", **self._extra_headers}
        req = urllib.request.Request(self._url, data=payload, headers=headers, method="POST")
        try:
            with urllib.request.urlopen(req, timeout=self._timeout) as response:
                status = response.status
                if status >= 400:
                    logger.warning(
                        "WebhookAlertHandler: webhook returned HTTP %d", status
                    )
                else:
                    logger.debug("WebhookAlertHandler: alert sent (HTTP %d)", status)
        except Exception:
            logger.exception(
                "WebhookAlertHandler: failed to send alert to %s", self._url
            )


class DriftAlertManager:
    """Evaluate :class:`DriftResult` objects and dispatch alerts to handlers.

    Parameters
    ----------
    handlers:
        A sequence of callables accepting a :class:`DriftAlert`.  Defaults to
        a single :class:`ConsoleAlertHandler`.
    min_severity:
        Minimum severity level to fire an alert for.  Options: ``"low"``,
        ``"medium"``, ``"high"``.
    cooldown_seconds:
        Suppress repeated alerts for the same agent within this window.
    """

    _SEVERITY_ORDER = {"none": 0, "low": 1, "medium": 2, "high": 3}

    def __init__(
        self,
        handlers: Optional[Sequence[AlertHandler]] = None,
        min_severity: str = "low",
        cooldown_seconds: float = 300.0,
    ) -> None:
        self._handlers: list[AlertHandler] = list(handlers or [ConsoleAlertHandler()])
        self._min_severity = min_severity
        self._cooldown = cooldown_seconds
        # {agent_id: last_alert_timestamp}
        self._last_alert: dict[str, float] = {}

    def add_handler(self, handler: AlertHandler) -> None:
        """Register an additional alert handler."""
        self._handlers.append(handler)

    def process(self, result: DriftResult) -> Optional[DriftAlert]:
        """Evaluate *result* and fire an alert if warranted.

        Returns
        -------
        The :class:`DriftAlert` that was fired, or ``None`` if no alert was sent.
        """
        if not result.drifted:
            return None

        min_rank = self._SEVERITY_ORDER.get(self._min_severity, 1)
        result_rank = self._SEVERITY_ORDER.get(result.severity, 0)
        if result_rank < min_rank:
            return None

        now = time.time()
        last = self._last_alert.get(result.agent_id, 0.0)
        if now - last < self._cooldown:
            logger.debug(
                "DriftAlertManager: suppressing alert for %s (cooldown)",
                result.agent_id,
            )
            return None

        alert = _result_to_alert(result)
        self._last_alert[result.agent_id] = now

        for handler in self._handlers:
            try:
                handler(alert)
            except Exception:
                logger.exception("DriftAlertManager: handler %s raised an exception", handler)

        return alert



"""CorrelationContext — trace context propagated across agent boundaries.

Carries the trace_id, parent_span_id, and baggage items that allow spans
from different agents to be linked into a coherent distributed trace.

Example
-------
::

    from agent_observability.correlation.correlation_context import CorrelationContext

    ctx = CorrelationContext.new_root()
    child_ctx = ctx.new_child_span("child-span-id")
    print(child_ctx.parent_span_id == ctx.span_id)  # True
"""
from __future__ import annotations

import uuid
from dataclasses import dataclass, field
from typing import Optional


@dataclass(frozen=True)
class BaggageItem:
    """A single baggage key-value pair propagated through a trace.

    Baggage items carry cross-cutting concerns like tenant ID, user ID,
    or feature flags across all spans in a trace.

    Attributes
    ----------
    key:
        Baggage item key. Must be non-empty.
    value:
        Baggage item value.
    """

    key: str
    value: str

    def __post_init__(self) -> None:
        if not self.key:
            raise ValueError("BaggageItem key must be non-empty.")

    def to_dict(self) -> dict[str, str]:
        """Serialise to a plain dictionary."""
        return {"key": self.key, "value": self.value}


@dataclass(frozen=True)
class CorrelationContext:
    """Trace context that propagates across agent boundaries.

    Carries a trace_id (shared across all spans in a trace), a span_id
    (unique to the current operation), and optional parent_span_id for
    linking parent and child spans.

    Attributes
    ----------
    trace_id:
        Globally unique trace identifier. Shared across all spans in one trace.
    span_id:
        Unique identifier for the current span.
    parent_span_id:
        ID of the parent span. None for root spans.
    baggage:
        Tuple of BaggageItem instances propagated through the trace.
    sampled:
        Whether this trace is being sampled for collection.
    """

    trace_id: str
    span_id: str
    parent_span_id: Optional[str] = None
    baggage: tuple[BaggageItem, ...] = field(default_factory=tuple)
    sampled: bool = True

    @classmethod
    def new_root(
        cls,
        *,
        trace_id: Optional[str] = None,
        sampled: bool = True,
    ) -> "CorrelationContext":
        """Create a new root context (no parent).

        Parameters
        ----------
        trace_id:
            Optional trace ID. Auto-generated if not provided.
        sampled:
            Whether this trace is sampled. Defaults to True.

        Returns
        -------
        CorrelationContext
            A new root context with no parent_span_id.
        """
        return cls(
            trace_id=trace_id or str(uuid.uuid4()),
            span_id=str(uuid.uuid4()),
            parent_span_id=None,
            sampled=sampled,
        )

    def new_child_span(
        self,
        span_id: Optional[str] = None,
    ) -> "CorrelationContext":
        """Create a child context inheriting trace_id and setting parent_span_id.

        Parameters
        ----------
        span_id:
            Optional span ID for the child. Auto-generated if not provided.

        Returns
        -------
        CorrelationContext
            A new child context with this span's span_id as parent_span_id.
        """
        return CorrelationContext(
            trace_id=self.trace_id,
            span_id=span_id or str(uuid.uuid4()),
            parent_span_id=self.span_id,
            baggage=self.baggage,
            sampled=self.sampled,
        )

    @property
    def is_root(self) -> bool:
        """True if this context has no parent span."""
        return self.parent_span_id is None

    def get_baggage(self, key: str) -> Optional[str]:
        """Return the value of a baggage item by key, or None if not found."""
        for item in self.baggage:
            if item.key == key:
                return item.value
        return None

    def with_baggage(self, key: str, value: str) -> "CorrelationContext":
        """Return a new context with an additional baggage item.

        If a baggage item with the same key already exists, it is replaced.

        Parameters
        ----------
        key:
            Baggage item key.
        value:
            Baggage item value.

        Returns
        -------
        CorrelationContext
            New context with the added/updated baggage item.
        """
        existing = tuple(item for item in self.baggage if item.key != key)
        new_baggage = existing + (BaggageItem(key=key, value=value),)
        return CorrelationContext(
            trace_id=self.trace_id,
            span_id=self.span_id,
            parent_span_id=self.parent_span_id,
            baggage=new_baggage,
            sampled=self.sampled,
        )

    def to_dict(self) -> dict[str, object]:
        """Serialise to a plain dictionary for header injection."""
        return {
            "trace_id": self.trace_id,
            "span_id": self.span_id,
            "parent_span_id": self.parent_span_id,
            "sampled": self.sampled,
            "baggage": [item.to_dict() for item in self.baggage],
        }

    @classmethod
    def from_dict(cls, data: dict[str, object]) -> "CorrelationContext":
        """Reconstruct a CorrelationContext from a header dictionary.

        Parameters
        ----------
        data:
            Dictionary as produced by :meth:`to_dict` or extracted from
            HTTP headers.

        Returns
        -------
        CorrelationContext
            Reconstructed context.

        Raises
        ------
        ValueError
            If required fields are missing.
        """
        trace_id = data.get("trace_id")
        span_id = data.get("span_id")
        if not trace_id or not span_id:
            raise ValueError("CorrelationContext requires 'trace_id' and 'span_id'.")

        raw_baggage = data.get("baggage", [])
        baggage_items: list[BaggageItem] = []
        if isinstance(raw_baggage, list):
            for item_dict in raw_baggage:
                if isinstance(item_dict, dict) and "key" in item_dict and "value" in item_dict:
                    baggage_items.append(
                        BaggageItem(key=str(item_dict["key"]), value=str(item_dict["value"]))
                    )

        return cls(
            trace_id=str(trace_id),
            span_id=str(span_id),
            parent_span_id=data.get("parent_span_id"),  # type: ignore[arg-type]
            baggage=tuple(baggage_items),
            sampled=bool(data.get("sampled", True)),
        )

    def to_headers(self) -> dict[str, str]:
        """Render as W3C traceparent-compatible headers.

        Returns
        -------
        dict[str, str]
            HTTP headers for trace context propagation.
        """
        headers: dict[str, str] = {}
        # W3C traceparent format: version-traceid-spanid-flags
        flags = "01" if self.sampled else "00"
        # Use first 32 hex chars of trace_id (strip dashes, pad if needed)
        trace_hex = self.trace_id.replace("-", "")[:32].ljust(32, "0")
        span_hex = self.span_id.replace("-", "")[:16].ljust(16, "0")
        headers["traceparent"] = f"00-{trace_hex}-{span_hex}-{flags}"

        if self.parent_span_id:
            headers["x-parent-span-id"] = self.parent_span_id

        if self.baggage:
            baggage_str = ",".join(f"{item.key}={item.value}" for item in self.baggage)
            headers["baggage"] = baggage_str

        return headers



"""Benchmark: Latency overhead of AgentTracer span creation vs no tracing.

Measures the wall-clock overhead introduced by creating and finishing an
AgentSpan relative to the equivalent no-op baseline. This is the primary
concern raised against tracing libraries.

Competitor context
------------------
- Langfuse issue #9429 (2024-09): users reported 2-4x latency overhead for
  synchronous SDK calls in tight loops.
  Source: https://github.com/langfuse/langfuse/issues/9429
- Langfuse issue #9766 (2024-10): follow-up confirming batch export reduces
  overhead to <5% in async usage.
  Source: https://github.com/langfuse/langfuse/issues/9766
- OpenLLMetry: no published overhead numbers; community reports <1ms per span.

agent-observability uses a NoOp fallback when OTel SDK is absent, making
the span lifecycle overhead pure Python object creation overhead.
"""
from __future__ import annotations

import json
import statistics
import sys
import time
from pathlib import Path

sys.path.insert(0, str(Path(__file__).parent))
sys.path.insert(0, str(Path(__file__).parent.parent / "src"))
sys.path.insert(0, str(Path(__file__).parent / "datasets"))

from agent_observability.spans.types import AgentTracer
from datasets.workload_generator import ActionKind, generate_workload


def _no_op_operation(index: int) -> str:
    """Simulate a trivially fast operation without any tracing overhead.

    Returns a string to prevent dead-code elimination by the interpreter.
    """
    return f"result_{index}"


def _traced_operation(tracer: AgentTracer, index: int) -> str:
    """The same trivial operation wrapped in an AgentTracer span."""
    with tracer.llm_call(model="gpt-4o-mini") as span:
        result = f"result_{index}"
        span.set_tokens(50, 100)
        span.set_cost(0.000015)
    return result


def run_benchmark(
    n_iterations: int = 2000,
    seed: int = 42,
) -> dict[str, object]:
    """Measure the latency overhead ratio of traced vs untraced operations.

    Parameters
    ----------
    n_iterations:
        Number of individual operations per condition.
    seed:
        Unused here (reserved for future parametric workloads).

    Returns
    -------
    dict with overhead metrics.
    """
    tracer = AgentTracer(
        tracer_name="bench-tracer",
        agent_id="bench-agent",
        session_id="bench-session",
    )

    # Warmup — 100 iterations of each to prime interpreter paths
    for w in range(100):
        _no_op_operation(w)
    for w in range(100):
        _traced_operation(tracer, w)

    # Baseline: untraced
    baseline_lats: list[float] = []
    for index in range(n_iterations):
        start = time.perf_counter()
        _no_op_operation(index)
        baseline_lats.append((time.perf_counter() - start) * 1_000_000)  # microseconds

    # Traced: same operation inside an AgentTracer span
    traced_lats: list[float] = []
    for index in range(n_iterations):
        start = time.perf_counter()
        _traced_operation(tracer, index)
        traced_lats.append((time.perf_counter() - start) * 1_000_000)

    def _percentile(data: list[float], pct: float) -> float:
        sorted_data = sorted(data)
        idx = int(len(sorted_data) * pct)
        idx = min(idx, len(sorted_data) - 1)
        return round(sorted_data[idx], 3)

    baseline_mean = statistics.mean(baseline_lats)
    traced_mean = statistics.mean(traced_lats)
    overhead_pct = (
        round((traced_mean - baseline_mean) / baseline_mean * 100, 2)
        if baseline_mean > 0
        else 0.0
    )

    return {
        "benchmark": "trace_overhead",
        "n_iterations": n_iterations,
        "baseline_us": {
            "p50": _percentile(baseline_lats, 0.50),
            "p95": _percentile(baseline_lats, 0.95),
            "p99": _percentile(baseline_lats, 0.99),
            "mean": round(baseline_mean, 3),
        },
        "traced_us": {
            "p50": _percentile(traced_lats, 0.50),
            "p95": _percentile(traced_lats, 0.95),
            "p99": _percentile(traced_lats, 0.99),
            "mean": round(traced_mean, 3),
        },
        "overhead_percent": overhead_pct,
        "note": (
            "Langfuse sync SDK showed 2-4x overhead (issues #9429, #9766). "
            "agent-observability uses NoOp spans when OTel SDK absent, "
            "so overhead is pure Python object lifecycle cost."
        ),
    }


if __name__ == "__main__":
    print("Running trace overhead benchmark (n=2000 iterations each)...")
    result = run_benchmark(n_iterations=2000)
    print(json.dumps(result, indent=2))
    results_dir = Path(__file__).parent / "results"
    results_dir.mkdir(exist_ok=True)
    output_path = results_dir / "baseline.json"
    with open(output_path, "w", encoding="utf-8") as fh:
        json.dump(result, fh, indent=2)
    print(f"\nResults saved to {output_path}")



"""Tests for OTel GenAI convention alignment."""
from __future__ import annotations

import pytest
from agent_observability.conventions.semantic import (
    LEGACY_TO_OTEL,
    AumOSAttributes,
    AumOSSpanExtension,
    GenAIAttributes,
    OTelGenAISpanName,
)


class TestOTelConventionMapping:
    """Verify legacy-to-OTel span type mapping."""

    def test_all_legacy_types_mapped(self) -> None:
        legacy_types = {
            "llm.call",
            "tool.invoke",
            "agent.delegate",
            "memory.read",
            "memory.write",
            "reasoning.step",
            "human.approval",
            "agent.error",
        }
        assert set(LEGACY_TO_OTEL.keys()) == legacy_types

    def test_official_otel_mappings(self) -> None:
        assert LEGACY_TO_OTEL["llm.call"] == "gen_ai.chat"
        assert LEGACY_TO_OTEL["tool.invoke"] == "gen_ai.execute_tool"
        assert LEGACY_TO_OTEL["agent.delegate"] == "invoke_agent"

    def test_extension_mappings_use_namespace(self) -> None:
        for key in [
            "memory.read",
            "memory.write",
            "reasoning.step",
            "human.approval",
            "agent.error",
        ]:
            assert LEGACY_TO_OTEL[key].startswith("gen_ai.agent.")

    def test_genai_attributes_match_spec(self) -> None:
        assert GenAIAttributes.SYSTEM == "gen_ai.system"
        assert GenAIAttributes.AGENT_NAME == "gen_ai.agent.name"
        assert GenAIAttributes.USAGE_INPUT_TOKENS == "gen_ai.usage.input_tokens"

    def test_otel_span_names_are_strings(self) -> None:
        for member in OTelGenAISpanName:
            assert isinstance(member.value, str)

    def test_aumos_extensions_are_strings(self) -> None:
        for member in AumOSSpanExtension:
            assert isinstance(member.value, str)

    def test_aumos_attributes_are_strings(self) -> None:
        for attr in [
            AumOSAttributes.COST_USD,
            AumOSAttributes.MEMORY_LAYER,
            AumOSAttributes.DRIFT_SCORE,
        ]:
            assert isinstance(attr, str)
            assert attr.startswith("aumos.")

    def test_legacy_mapping_is_complete(self) -> None:
        """Every legacy span type should have a non-empty OTel mapping."""
        for legacy, otel in LEGACY_TO_OTEL.items():
            assert otel, f"Empty OTel mapping for legacy span type: {legacy}"

    def test_no_duplicate_otel_values(self) -> None:
        """No two legacy types should map to the same OTel convention."""
        values = list(LEGACY_TO_OTEL.values())
        assert len(values) == len(set(values)), "Duplicate OTel mappings found"


class TestOTelSpanNameEnum:
    """Verify OTelGenAISpanName enum values match the spec."""

    def test_chat_span_name(self) -> None:
        assert OTelGenAISpanName.CHAT == "gen_ai.chat"
        assert OTelGenAISpanName.CHAT.value == "gen_ai.chat"

    def test_execute_tool_span_name(self) -> None:
        assert OTelGenAISpanName.EXECUTE_TOOL == "gen_ai.execute_tool"

    def test_invoke_agent_span_name(self) -> None:
        assert OTelGenAISpanName.INVOKE_AGENT == "invoke_agent"

    def test_enum_members_count(self) -> None:
        assert len(OTelGenAISpanName) == 3

    def test_enum_is_str_subclass(self) -> None:
        assert issubclass(OTelGenAISpanName, str)

    def test_enum_string_comparison(self) -> None:
        """str(Enum) should equal the value for str-based enums."""
        assert OTelGenAISpanName.CHAT == OTelGenAISpanName.CHAT.value


class TestAumOSSpanExtensionEnum:
    """Verify AumOSSpanExtension enum values use gen_ai.agent namespace."""

    def test_memory_read(self) -> None:
        assert AumOSSpanExtension.MEMORY_READ == "gen_ai.agent.memory.read"

    def test_memory_write(self) -> None:
        assert AumOSSpanExtension.MEMORY_WRITE == "gen_ai.agent.memory.write"

    def test_reasoning_step(self) -> None:
        assert AumOSSpanExtension.REASONING_STEP == "gen_ai.agent.reasoning.step"

    def test_human_approval(self) -> None:
        assert AumOSSpanExtension.HUMAN_APPROVAL == "gen_ai.agent.human_approval"

    def test_agent_error(self) -> None:
        assert AumOSSpanExtension.AGENT_ERROR == "gen_ai.agent.error"

    def test_all_extensions_use_agent_namespace(self) -> None:
        for member in AumOSSpanExtension:
            assert member.value.startswith("gen_ai.agent."), (
                f"{member.name} does not use gen_ai.agent namespace: {member.value}"
            )

    def test_enum_members_count(self) -> None:
        assert len(AumOSSpanExtension) == 5


class TestGenAIAttributes:
    """Verify GenAIAttributes class constants match the OTel spec."""

    def test_system(self) -> None:
        assert GenAIAttributes.SYSTEM == "gen_ai.system"

    def test_request_model(self) -> None:
        assert GenAIAttributes.REQUEST_MODEL == "gen_ai.request.model"

    def test_request_max_tokens(self) -> None:
        assert GenAIAttributes.REQUEST_MAX_TOKENS == "gen_ai.request.max_tokens"

    def test_request_temperature(self) -> None:
        assert GenAIAttributes.REQUEST_TEMPERATURE == "gen_ai.request.temperature"

    def test_response_model(self) -> None:
        assert GenAIAttributes.RESPONSE_MODEL == "gen_ai.response.model"

    def test_response_finish_reasons(self) -> None:
        assert GenAIAttributes.RESPONSE_FINISH_REASONS == "gen_ai.response.finish_reasons"

    def test_usage_input_tokens(self) -> None:
        assert GenAIAttributes.USAGE_INPUT_TOKENS == "gen_ai.usage.input_tokens"

    def test_usage_output_tokens(self) -> None:
        assert GenAIAttributes.USAGE_OUTPUT_TOKENS == "gen_ai.usage.output_tokens"

    def test_agent_name(self) -> None:
        assert GenAIAttributes.AGENT_NAME == "gen_ai.agent.name"

    def test_agent_description(self) -> None:
        assert GenAIAttributes.AGENT_DESCRIPTION == "gen_ai.agent.description"

    def test_tool_name(self) -> None:
        assert GenAIAttributes.TOOL_NAME == "gen_ai.tool.name"

    def test_tool_description(self) -> None:
        assert GenAIAttributes.TOOL_DESCRIPTION == "gen_ai.tool.description"

    def test_all_attributes_use_gen_ai_prefix(self) -> None:
        attrs = [
            name
            for name in dir(GenAIAttributes)
            if not name.startswith("_")
        ]
        for attr_name in attrs:
            value = getattr(GenAIAttributes, attr_name)
            assert isinstance(value, str), f"{attr_name} is not a string"
            assert value.startswith("gen_ai."), f"{attr_name} = {value!r} missing gen_ai. prefix"


class TestAumOSAttributes:
    """Verify AumOSAttributes class constants use the aumos namespace."""

    def test_cost_usd(self) -> None:
        assert AumOSAttributes.COST_USD == "aumos.cost.usd"

    def test_memory_layer(self) -> None:
        assert AumOSAttributes.MEMORY_LAYER == "aumos.memory.layer"

    def test_memory_key(self) -> None:
        assert AumOSAttributes.MEMORY_KEY == "aumos.memory.key"

    def test_reasoning_confidence(self) -> None:
        assert AumOSAttributes.REASONING_CONFIDENCE == "aumos.reasoning.confidence"

    def test_approval_required(self) -> None:
        assert AumOSAttributes.APPROVAL_REQUIRED == "aumos.approval.required"

    def test_approval_granted(self) -> None:
        assert AumOSAttributes.APPROVAL_GRANTED == "aumos.approval.granted"

    def test_error_category(self) -> None:
        assert AumOSAttributes.ERROR_CATEGORY == "aumos.error.category"

    def test_drift_score(self) -> None:
        assert AumOSAttributes.DRIFT_SCORE == "aumos.drift.score"

    def test_all_attributes_use_aumos_prefix(self) -> None:
        attrs = [name for name in dir(AumOSAttributes) if not name.startswith("_")]
        for attr_name in attrs:
            value = getattr(AumOSAttributes, attr_name)
            assert isinstance(value, str), f"{attr_name} is not a string"
            assert value.startswith("aumos."), f"{attr_name} = {value!r} missing aumos. prefix"



"""Comprehensive tests for the drift detection subsystem.

Covers:
- drift.features  (SpanRecord, BehavioralFeatures, FeatureExtractor)
- drift.baseline  (BaselineStats, AgentBaseline, BaselineComputer)
- drift.detector  (DriftResult, DriftDetector)
- drift.history   (BaselineHistory)
- drift.metric_drift (MetricDriftDetector, DriftReport, DriftSeverity)
- drift.alerts    (DriftAlert, ConsoleAlertHandler, WebhookAlertHandler, DriftAlertManager)
- drift.__init__  (public re-exports)
"""
from __future__ import annotations

import json
import time
import threading
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock, patch

import pytest

from agent_observability.drift.features import (
    BehavioralFeatures,
    FeatureExtractor,
    SpanRecord,
)
from agent_observability.drift.baseline import (
    AgentBaseline,
    BaselineComputer,
    BaselineStats,
    _mean,
    _std_dev,
)
from agent_observability.drift.detector import DriftDetector, DriftResult
from agent_observability.drift.history import BaselineHistory
from agent_observability.drift.metric_drift import (
    DriftReport,
    DriftSeverity,
    MetricDriftDetector,
)
from agent_observability.drift.alerts import (
    ConsoleAlertHandler,
    DriftAlert,
    DriftAlertManager,
    WebhookAlertHandler,
    _result_to_alert,
)
from agent_observability.drift import (
    AgentBaseline as PublicBaseline,
    DriftDetector as PublicDetector,
    MetricDriftDetector as PublicMetricDetector,
)


# ── Fixtures ───────────────────────────────────────────────────────────────────


def _make_span(
    span_kind: str = "llm_call",
    duration_ms: float = 100.0,
    input_tokens: int = 50,
    output_tokens: int = 25,
    cost_usd: float = 0.001,
    tool_name: str = "",
    tool_success: bool = True,
    error: bool = False,
) -> SpanRecord:
    return SpanRecord(
        span_kind=span_kind,
        duration_ms=duration_ms,
        input_tokens=input_tokens,
        output_tokens=output_tokens,
        cost_usd=cost_usd,
        tool_name=tool_name,
        tool_success=tool_success,
        error=error,
    )


def _make_baseline(
    agent_id: str = "agent-1",
    means: dict[str, float] | None = None,
    std_devs: dict[str, float] | None = None,
) -> AgentBaseline:
    means = means or {"span_count": 10.0, "mean_duration_ms": 100.0}
    std_devs = std_devs or {"span_count": 1.0, "mean_duration_ms": 10.0}
    feature_stats = {
        key: BaselineStats(mean=means[key], std_dev=std_devs[key], sample_count=5)
        for key in means
    }
    return AgentBaseline(
        agent_id=agent_id,
        created_at=time.time() - 60,
        span_count=50,
        feature_stats=feature_stats,
    )


def _make_drift_result(
    drifted: bool = True,
    severity: str = "high",
    max_z_score: float = 5.0,
    agent_id: str = "agent-1",
) -> DriftResult:
    return DriftResult(
        agent_id=agent_id,
        timestamp=time.time(),
        drifted=drifted,
        max_z_score=max_z_score,
        threshold=3.0,
        drifted_features={"span_count": max_z_score} if drifted else {},
        all_z_scores={"span_count": max_z_score},
        baseline_age_seconds=60.0,
        window_span_count=20,
    )


# ── SpanRecord ─────────────────────────────────────────────────────────────────


class TestSpanRecord:
    def test_defaults(self) -> None:
        span = SpanRecord(span_kind="llm_call", duration_ms=50.0)
        assert span.input_tokens == 0
        assert span.output_tokens == 0
        assert span.cost_usd == 0.0
        assert span.tool_name == ""
        assert span.tool_success is True
        assert span.error is False
        assert span.attributes == {}

    def test_from_dict_with_full_attributes(self) -> None:
        data = {
            "start_time_ns": 1_000_000_000,
            "end_time_ns": 1_100_000_000,
            "status": "OK",
            "attributes": {
                "agent.span.kind": "tool_invoke",
                "llm.tokens.input": 100,
                "llm.tokens.output": 50,
                "llm.cost.usd": 0.005,
                "tool.name": "web_search",
                "tool.success": True,
            },
        }
        span = SpanRecord.from_dict(data)
        assert span.span_kind == "tool_invoke"
        assert span.duration_ms == pytest.approx(100.0)
        assert span.input_tokens == 100
        assert span.output_tokens == 50
        assert span.cost_usd == pytest.approx(0.005)
        assert span.tool_name == "web_search"
        assert span.tool_success is True
        assert span.error is False

    def test_from_dict_error_status(self) -> None:
        data = {"status": "ERROR", "attributes": {}}
        span = SpanRecord.from_dict(data)
        assert span.error is True

    def test_from_dict_missing_times(self) -> None:
        data = {"attributes": {"agent.span.kind": "reasoning_step"}}
        span = SpanRecord.from_dict(data)
        assert span.duration_ms == 0.0

    def test_from_dict_zero_times(self) -> None:
        data = {"start_time_ns": 0, "end_time_ns": 0, "attributes": {}}
        span = SpanRecord.from_dict(data)
        assert span.duration_ms == 0.0

    def test_from_dict_invalid_attributes_type(self) -> None:
        data = {"attributes": "not_a_dict"}
        span = SpanRecord.from_dict(data)
        assert span.span_kind == ""

    def test_from_dict_uses_span_kind_fallback(self) -> None:
        data = {"span_kind": "memory_read", "attributes": {}}
        span = SpanRecord.from_dict(data)
        assert span.span_kind == "memory_read"


# ── BehavioralFeatures ─────────────────────────────────────────────────────────


class TestBehavioralFeatures:
    def test_to_vector_basic_fields(self) -> None:
        features = BehavioralFeatures(
            span_count=10,
            mean_duration_ms=100.0,
            p95_duration_ms=200.0,
            mean_input_tokens=50.0,
            mean_output_tokens=25.0,
            mean_cost_usd=0.001,
            total_cost_usd=0.01,
            error_rate=0.1,
            tool_failure_rate=0.05,
        )
        vec = features.to_vector()
        assert vec["span_count"] == 10.0
        assert vec["mean_duration_ms"] == 100.0
        assert vec["p95_duration_ms"] == 200.0
        assert vec["error_rate"] == 0.1

    def test_to_vector_includes_kind_fractions(self) -> None:
        features = BehavioralFeatures(
            kind_fractions={"llm_call": 0.6, "tool_invoke": 0.4},
        )
        vec = features.to_vector()
        assert vec["kind.llm_call"] == 0.6
        assert vec["kind.tool_invoke"] == 0.4

    def test_to_vector_includes_tool_fractions(self) -> None:
        features = BehavioralFeatures(
            tool_fractions={"web_search": 0.7, "code_exec": 0.3},
        )
        vec = features.to_vector()
        assert vec["tool.web_search"] == 0.7
        assert vec["tool.code_exec"] == 0.3

    def test_to_vector_empty_fractions(self) -> None:
        features = BehavioralFeatures()
        vec = features.to_vector()
        assert "span_count" in vec
        assert not any(k.startswith("kind.") for k in vec)
        assert not any(k.startswith("tool.") for k in vec)


# ── FeatureExtractor ───────────────────────────────────────────────────────────


class TestFeatureExtractor:
    def setup_method(self) -> None:
        self.extractor = FeatureExtractor()

    def test_empty_spans(self) -> None:
        result = self.extractor.extract([])
        assert result.span_count == 0
        assert result.mean_duration_ms == 0.0

    def test_single_span(self) -> None:
        spans = [_make_span(duration_ms=100.0, input_tokens=10, output_tokens=5, cost_usd=0.001)]
        result = self.extractor.extract(spans)
        assert result.span_count == 1
        assert result.mean_duration_ms == pytest.approx(100.0)
        assert result.mean_input_tokens == pytest.approx(10.0)
        assert result.mean_output_tokens == pytest.approx(5.0)
        assert result.mean_cost_usd == pytest.approx(0.001)
        assert result.total_cost_usd == pytest.approx(0.001)

    def test_error_rate(self) -> None:
        spans = [
            _make_span(error=True),
            _make_span(error=True),
            _make_span(error=False),
            _make_span(error=False),
        ]
        result = self.extractor.extract(spans)
        assert result.error_rate == pytest.approx(0.5)

    def test_tool_failure_rate(self) -> None:
        spans = [
            _make_span(span_kind="tool_invoke", tool_name="search", tool_success=False),
            _make_span(span_kind="tool_invoke", tool_name="search", tool_success=True),
            _make_span(span_kind="llm_call"),
        ]
        result = self.extractor.extract(spans)
        # 1 failure out of 2 tool spans
        assert result.tool_failure_rate == pytest.approx(0.5)

    def test_kind_fractions(self) -> None:
        spans = [
            _make_span(span_kind="llm_call"),
            _make_span(span_kind="llm_call"),
            _make_span(span_kind="tool_invoke"),
        ]
        result = self.extractor.extract(spans)
        assert result.kind_fractions["llm_call"] == pytest.approx(2 / 3)
        assert result.kind_fractions["tool_invoke"] == pytest.approx(1 / 3)

    def test_tool_fractions(self) -> None:
        spans = [
            _make_span(span_kind="tool_invoke", tool_name="search"),
            _make_span(span_kind="tool_invoke", tool_name="search"),
            _make_span(span_kind="tool_invoke", tool_name="code"),
        ]
        result = self.extractor.extract(spans)
        assert result.tool_fractions["search"] == pytest.approx(2 / 3)
        assert result.tool_fractions["code"] == pytest.approx(1 / 3)

    def test_p95_duration(self) -> None:
        spans = [_make_span(duration_ms=float(i)) for i in range(1, 21)]
        result = self.extractor.extract(spans)
        # p95 of 20 values = index ceil(0.95*20)-1 = 18 in sorted array
        assert result.p95_duration_ms == pytest.approx(19.0)

    def test_tool_span_without_tool_name_ignored_in_fractions(self) -> None:
        spans = [_make_span(span_kind="tool_invoke", tool_name="")]
        result = self.extractor.extract(spans)
        assert result.tool_fractions == {}


# ── BaselineStats ──────────────────────────────────────────────────────────────


class TestBaselineStats:
    def test_z_score_zero_std(self) -> None:
        stats = BaselineStats(mean=10.0, std_dev=0.0, sample_count=5)
        assert stats.z_score(10.0) == 0.0
        assert stats.z_score(999.0) == 0.0

    def test_z_score_positive(self) -> None:
        stats = BaselineStats(mean=100.0, std_dev=10.0, sample_count=10)
        assert stats.z_score(120.0) == pytest.approx(2.0)

    def test_z_score_negative(self) -> None:
        stats = BaselineStats(mean=100.0, std_dev=10.0, sample_count=10)
        assert stats.z_score(80.0) == pytest.approx(-2.0)


# ── AgentBaseline ──────────────────────────────────────────────────────────────


class TestAgentBaseline:
    def test_z_scores_returns_scores_for_known_features(self) -> None:
        baseline = _make_baseline(
            means={"span_count": 10.0, "mean_duration_ms": 100.0},
            std_devs={"span_count": 1.0, "mean_duration_ms": 10.0},
        )
        features = BehavioralFeatures(span_count=13, mean_duration_ms=100.0)
        scores = baseline.z_scores(features)
        assert scores["span_count"] == pytest.approx(3.0)
        assert scores["mean_duration_ms"] == pytest.approx(0.0)

    def test_z_scores_ignores_unknown_features(self) -> None:
        baseline = _make_baseline(means={"span_count": 10.0}, std_devs={"span_count": 1.0})
        features = BehavioralFeatures(span_count=10, mean_duration_ms=999.0)
        scores = baseline.z_scores(features)
        assert "mean_duration_ms" not in scores

    def test_max_z_score_no_features(self) -> None:
        baseline = AgentBaseline(
            agent_id="a", created_at=time.time(), span_count=0, feature_stats={}
        )
        features = BehavioralFeatures()
        assert baseline.max_z_score(features) == 0.0

    def test_max_z_score_picks_largest_abs(self) -> None:
        baseline = _make_baseline(
            means={"span_count": 10.0, "mean_duration_ms": 100.0},
            std_devs={"span_count": 1.0, "mean_duration_ms": 10.0},
        )
        features = BehavioralFeatures(span_count=14, mean_duration_ms=80.0)
        max_z = baseline.max_z_score(features)
        # span_count z=4.0, duration z=-2.0 → max abs = 4.0
        assert max_z == pytest.approx(4.0)


# ── BaselineComputer ───────────────────────────────────────────────────────────


class TestBaselineComputer:
    def setup_method(self) -> None:
        self.computer = BaselineComputer()

    def test_raises_on_empty_windows(self) -> None:
        with pytest.raises(ValueError, match="At least one span window"):
            self.computer.compute("agent-1", [])

    def test_single_window(self) -> None:
        spans = [_make_span(duration_ms=100.0) for _ in range(5)]
        baseline = self.computer.compute("agent-1", [spans])
        assert baseline.agent_id == "agent-1"
        assert baseline.span_count == 5
        assert "mean_duration_ms" in baseline.feature_stats

    def test_multiple_windows_averages_stats(self) -> None:
        window_a = [_make_span(duration_ms=100.0) for _ in range(5)]
        window_b = [_make_span(duration_ms=200.0) for _ in range(5)]
        baseline = self.computer.compute("agent-1", [window_a, window_b])
        assert baseline.span_count == 10
        # Mean of [100, 200] = 150
        mean = baseline.feature_stats["mean_duration_ms"].mean
        assert mean == pytest.approx(150.0)

    def test_metadata_attached(self) -> None:
        spans = [_make_span() for _ in range(3)]
        meta = {"env": "prod", "version": "1.2"}
        baseline = self.computer.compute("agent-1", [spans], metadata=meta)
        assert baseline.metadata == meta

    def test_single_window_std_is_zero(self) -> None:
        spans = [_make_span(duration_ms=100.0) for _ in range(5)]
        baseline = self.computer.compute("agent-1", [spans])
        # With one window, std_dev should be 0
        assert baseline.feature_stats["mean_duration_ms"].std_dev == 0.0

    def test_custom_extractor(self) -> None:
        extractor = FeatureExtractor()
        computer = BaselineComputer(extractor=extractor)
        spans = [_make_span() for _ in range(2)]
        baseline = computer.compute("agent-x", [spans])
        assert baseline.agent_id == "agent-x"


# ── _mean and _std_dev utilities ───────────────────────────────────────────────


class TestBaselineUtilities:
    def test_mean_empty(self) -> None:
        assert _mean([]) == 0.0

    def test_mean_values(self) -> None:
        assert _mean([1.0, 2.0, 3.0]) == pytest.approx(2.0)

    def test_std_dev_single_value(self) -> None:
        assert _std_dev([5.0], 5.0) == 0.0

    def test_std_dev_multiple(self) -> None:
        values = [2.0, 4.0, 4.0, 4.0, 5.0, 5.0, 7.0, 9.0]
        mean_val = sum(values) / len(values)
        std = _std_dev(values, mean_val)
        assert std > 0.0


# ── DriftResult ────────────────────────────────────────────────────────────────


class TestDriftResult:
    def test_severity_none(self) -> None:
        result = _make_drift_result(drifted=False, max_z_score=0.0)
        assert result.severity == "none"

    def test_severity_low(self) -> None:
        # ratio < 1.5 → low
        result = _make_drift_result(max_z_score=3.0 * 1.2)  # ratio = 1.2 → low
        assert result.severity == "low"

    def test_severity_medium(self) -> None:
        # ratio between 1.5 and 2.5
        result = _make_drift_result(max_z_score=3.0 * 2.0)  # ratio = 2.0
        assert result.severity == "medium"

    def test_severity_high(self) -> None:
        result = _make_drift_result(max_z_score=3.0 * 3.0)  # ratio = 3.0
        assert result.severity == "high"


# ── DriftDetector ──────────────────────────────────────────────────────────────


class TestDriftDetector:
    def test_raises_on_non_positive_threshold(self) -> None:
        with pytest.raises(ValueError, match="sigma_threshold must be positive"):
            DriftDetector(sigma_threshold=0)

    def test_skips_small_window(self) -> None:
        detector = DriftDetector(min_window_spans=5)
        baseline = _make_baseline()
        result = detector.check(baseline, [_make_span()])
        assert result.drifted is False
        assert "minimum" in result.notes.lower()

    def test_healthy_result(self) -> None:
        computer = BaselineComputer()
        windows = [[_make_span(duration_ms=100.0) for _ in range(10)] for _ in range(5)]
        baseline = computer.compute("agent-1", windows)
        detector = DriftDetector(sigma_threshold=3.0, min_window_spans=10)
        # Same pattern — should not drift
        current = [_make_span(duration_ms=100.0) for _ in range(10)]
        result = detector.check(baseline, current)
        assert result.agent_id == "agent-1"
        assert isinstance(result.drifted, bool)

    def test_drifted_result(self) -> None:
        """Force drift by setting a tiny std_dev and a very different current window."""
        baseline = AgentBaseline(
            agent_id="agent-drift",
            created_at=time.time() - 3600,
            span_count=100,
            feature_stats={
                "mean_duration_ms": BaselineStats(mean=100.0, std_dev=1.0, sample_count=50),
                "span_count": BaselineStats(mean=10.0, std_dev=0.5, sample_count=50),
                "mean_cost_usd": BaselineStats(mean=0.001, std_dev=0.0001, sample_count=50),
                "total_cost_usd": BaselineStats(mean=0.01, std_dev=0.001, sample_count=50),
                "mean_input_tokens": BaselineStats(mean=50.0, std_dev=1.0, sample_count=50),
                "mean_output_tokens": BaselineStats(mean=25.0, std_dev=1.0, sample_count=50),
                "error_rate": BaselineStats(mean=0.0, std_dev=0.01, sample_count=50),
                "tool_failure_rate": BaselineStats(mean=0.0, std_dev=0.01, sample_count=50),
                "p95_duration_ms": BaselineStats(mean=150.0, std_dev=2.0, sample_count=50),
            },
        )
        detector = DriftDetector(sigma_threshold=3.0, min_window_spans=5)
        # Massive deviation in duration
        drifted_spans = [_make_span(duration_ms=900.0) for _ in range(10)]
        result = detector.check(baseline, drifted_spans)
        assert result.drifted is True
        assert result.max_z_score > 3.0

    def test_check_features_uses_pre_extracted(self) -> None:
        baseline = _make_baseline(
            means={"span_count": 10.0},
            std_devs={"span_count": 1.0},
        )
        detector = DriftDetector(sigma_threshold=3.0)
        features = BehavioralFeatures(span_count=10)
        result = detector.check_features(baseline, features)
        assert result.agent_id == "agent-1"
        assert result.drifted is False

    def test_window_span_count_recorded(self) -> None:
        baseline = _make_baseline()
        detector = DriftDetector(min_window_spans=5)
        current = [_make_span() for _ in range(12)]
        result = detector.check(baseline, current)
        assert result.window_span_count == 12


# ── BaselineHistory ────────────────────────────────────────────────────────────


class TestBaselineHistory:
    def test_save_and_latest(self) -> None:
        history = BaselineHistory()
        baseline = _make_baseline("agent-a")
        history.save(baseline)
        latest = history.latest("agent-a")
        assert latest is not None
        assert latest.agent_id == "agent-a"

    def test_latest_returns_none_for_unknown_agent(self) -> None:
        history = BaselineHistory()
        assert history.latest("unknown") is None

    def test_all_for_agent(self) -> None:
        history = BaselineHistory()
        b1 = _make_baseline("agent-b")
        b2 = _make_baseline("agent-b")
        history.save(b1)
        history.save(b2)
        all_b = history.all_for_agent("agent-b")
        assert len(all_b) == 2

    def test_all_agent_ids(self) -> None:
        history = BaselineHistory()
        history.save(_make_baseline("agent-x"))
        history.save(_make_baseline("agent-y"))
        ids = history.all_agent_ids()
        assert ids == ["agent-x", "agent-y"]

    def test_delete_agent(self) -> None:
        history = BaselineHistory()
        history.save(_make_baseline("agent-del"))
        history.delete_agent("agent-del")
        assert history.latest("agent-del") is None

    def test_max_per_agent_prunes_oldest(self) -> None:
        history = BaselineHistory(max_per_agent=3)
        for _ in range(5):
            history.save(_make_baseline("agent-max"))
        all_b = history.all_for_agent("agent-max")
        assert len(all_b) == 3

    def test_thread_safety(self) -> None:
        history = BaselineHistory()
        errors: list[Exception] = []

        def worker(agent_id: str) -> None:
            try:
                for _ in range(10):
                    history.save(_make_baseline(agent_id))
            except Exception as exc:
                errors.append(exc)

        threads = [threading.Thread(target=worker, args=(f"agent-{i}",)) for i in range(5)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert errors == []

    def test_persist_and_load(self, tmp_path: Path) -> None:
        persist_file = str(tmp_path / "baselines.json")
        history = BaselineHistory(persist_path=persist_file)
        baseline = _make_baseline("agent-persist")
        history.save(baseline)

        # Verify file was written
        assert Path(persist_file).exists()

        # Load into new instance
        history2 = BaselineHistory(persist_path=persist_file)
        loaded = history2.latest("agent-persist")
        assert loaded is not None
        assert loaded.agent_id == "agent-persist"

    def test_persist_and_delete(self, tmp_path: Path) -> None:
        persist_file = str(tmp_path / "baselines.json")
        history = BaselineHistory(persist_path=persist_file)
        history.save(_make_baseline("agent-del-persist"))
        history.delete_agent("agent-del-persist")
        # Should still write without error
        assert Path(persist_file).exists()

    def test_load_corrupt_file_logs_and_skips(self, tmp_path: Path) -> None:
        persist_file = str(tmp_path / "corrupt.json")
        Path(persist_file).write_text("not valid json", encoding="utf-8")
        # Should not raise — just log
        history = BaselineHistory(persist_path=persist_file)
        assert history.all_agent_ids() == []


# ── MetricDriftDetector ────────────────────────────────────────────────────────


class TestMetricDriftDetector:
    def setup_method(self) -> None:
        self.detector = MetricDriftDetector()

    def _fill_baseline(self, agent_id: str, metric: str, values: list[float]) -> None:
        for v in values:
            self.detector.record_metric(agent_id, metric, v)

    def test_no_drift_below_min_samples(self) -> None:
        for i in range(9):
            self.detector.record_metric("agent-1", "latency_ms", float(i))
        report = self.detector.check_drift("agent-1", "latency_ms", 999.0)
        assert report.drifted is False
        assert report.severity == DriftSeverity.NONE

    def test_no_drift_within_threshold(self) -> None:
        # Use a varied baseline so stddev > 0, then check a value within normal range
        values = [100.0 + (i % 5) for i in range(20)]  # values 100-104, repeated
        self._fill_baseline("agent-1", "latency", values)
        report = self.detector.check_drift("agent-1", "latency", 102.0)
        assert report.drifted is False

    def test_low_drift(self) -> None:
        # Mean=100, std~0.0 from constant values won't trigger; use varied baseline
        self._fill_baseline("agent-1", "latency", [90.0, 95.0, 100.0, 105.0, 110.0] * 4)
        report = self.detector.check_drift("agent-1", "latency", 120.0)
        assert report.drifted is True

    def test_high_drift(self) -> None:
        self._fill_baseline("agent-2", "cost", [10.0] * 20)
        report = self.detector.check_drift("agent-2", "cost", 50.0)
        # stddev of constant series = 0 → z = inf → critical
        assert report.drifted is True
        assert report.severity == DriftSeverity.CRITICAL

    def test_stable_baseline_exact_match_no_drift(self) -> None:
        self._fill_baseline("agent-3", "cost", [10.0] * 20)
        report = self.detector.check_drift("agent-3", "cost", 10.0)
        assert report.drifted is False

    def test_get_baseline_insufficient_samples(self) -> None:
        mean, stddev = self.detector.get_baseline("agent-x", "nonexistent")
        assert mean == 0.0
        assert stddev == 0.0

    def test_get_baseline_sufficient_samples(self) -> None:
        self._fill_baseline("agent-4", "tokens", [100.0] * 10)
        mean, stddev = self.detector.get_baseline("agent-4", "tokens")
        assert mean == pytest.approx(100.0)
        assert stddev == pytest.approx(0.0)

    def test_check_all(self) -> None:
        self._fill_baseline("agent-5", "latency", [100.0] * 15)
        self._fill_baseline("agent-5", "cost", [0.01] * 15)
        reports = self.detector.check_all("agent-5", {"latency": 100.0, "cost": 0.01})
        assert len(reports) == 2
        assert all(isinstance(r, DriftReport) for r in reports)

    def test_custom_thresholds(self) -> None:
        custom = {
            DriftSeverity.LOW: 1.0,
            DriftSeverity.MEDIUM: 2.0,
            DriftSeverity.HIGH: 3.0,
            DriftSeverity.CRITICAL: 5.0,
        }
        detector = MetricDriftDetector(z_score_thresholds=custom)
        assert detector.z_score_thresholds[DriftSeverity.LOW] == 1.0

    def test_classify_severity_boundaries(self) -> None:
        assert self.detector._classify_severity(0.0) == DriftSeverity.NONE
        assert self.detector._classify_severity(1.5) == DriftSeverity.LOW
        assert self.detector._classify_severity(2.0) == DriftSeverity.MEDIUM
        assert self.detector._classify_severity(3.0) == DriftSeverity.HIGH
        assert self.detector._classify_severity(4.0) == DriftSeverity.CRITICAL


# ── DriftSeverity ──────────────────────────────────────────────────────────────


class TestDriftSeverity:
    def test_values(self) -> None:
        assert DriftSeverity.NONE.value == "none"
        assert DriftSeverity.LOW.value == "low"
        assert DriftSeverity.MEDIUM.value == "medium"
        assert DriftSeverity.HIGH.value == "high"
        assert DriftSeverity.CRITICAL.value == "critical"


# ── DriftAlert ─────────────────────────────────────────────────────────────────


class TestDriftAlert:
    def test_to_dict_round_trip(self) -> None:
        alert = DriftAlert(
            agent_id="agent-1",
            severity="high",
            max_z_score=5.0,
            threshold=3.0,
            drifted_features={"span_count": 5.0},
            baseline_age_seconds=60.0,
            window_span_count=20,
            timestamp=1234567890.0,
            message="test alert",
        )
        d = alert.to_dict()
        assert d["agent_id"] == "agent-1"
        assert d["severity"] == "high"
        assert d["message"] == "test alert"
        assert "drifted_features" in d


# ── _result_to_alert ───────────────────────────────────────────────────────────


class TestResultToAlert:
    def test_conversion(self) -> None:
        result = _make_drift_result(drifted=True, max_z_score=5.0)
        result.drifted_features = {"feat_a": 5.0, "feat_b": 4.0, "feat_c": 3.0, "feat_d": 2.0}
        alert = _result_to_alert(result)
        assert alert.agent_id == result.agent_id
        assert alert.severity == result.severity
        assert "feat_a" in alert.message

    def test_no_drifted_features_message(self) -> None:
        result = _make_drift_result(drifted=True, max_z_score=5.0)
        result.drifted_features = {}
        alert = _result_to_alert(result)
        assert "none" in alert.message.lower()


# ── ConsoleAlertHandler ────────────────────────────────────────────────────────


class TestConsoleAlertHandler:
    def _make_alert(self, severity: str) -> DriftAlert:
        return DriftAlert(
            agent_id="agent-1",
            severity=severity,
            max_z_score=5.0,
            threshold=3.0,
            drifted_features={},
            baseline_age_seconds=60.0,
            window_span_count=10,
            timestamp=time.time(),
            message="test",
        )

    def test_low_severity_logs_info(self) -> None:
        handler = ConsoleAlertHandler()
        with patch("agent_observability.drift.alerts.logger") as mock_logger:
            handler(self._make_alert("low"))
            mock_logger.info.assert_called_once()

    def test_medium_severity_logs_warning(self) -> None:
        handler = ConsoleAlertHandler()
        with patch("agent_observability.drift.alerts.logger") as mock_logger:
            handler(self._make_alert("medium"))
            mock_logger.warning.assert_called_once()

    def test_high_severity_logs_error(self) -> None:
        handler = ConsoleAlertHandler()
        with patch("agent_observability.drift.alerts.logger") as mock_logger:
            handler(self._make_alert("high"))
            mock_logger.error.assert_called_once()

    def test_unknown_severity_falls_back_to_warning(self) -> None:
        handler = ConsoleAlertHandler()
        with patch("agent_observability.drift.alerts.logger") as mock_logger:
            handler(self._make_alert("critical"))
            mock_logger.warning.assert_called_once()


# ── WebhookAlertHandler ────────────────────────────────────────────────────────


class TestWebhookAlertHandler:
    def _make_alert(self) -> DriftAlert:
        return DriftAlert(
            agent_id="agent-1",
            severity="high",
            max_z_score=5.0,
            threshold=3.0,
            drifted_features={},
            baseline_age_seconds=60.0,
            window_span_count=10,
            timestamp=time.time(),
            message="test",
        )

    def test_successful_post(self) -> None:
        handler = WebhookAlertHandler("http://example.com/webhook", timeout_seconds=5.0)
        mock_response = MagicMock()
        mock_response.__enter__ = lambda s: s
        mock_response.__exit__ = MagicMock(return_value=False)
        mock_response.status = 200

        with patch("urllib.request.urlopen", return_value=mock_response):
            handler(self._make_alert())

    def test_http_error_status_logs_warning(self) -> None:
        handler = WebhookAlertHandler("http://example.com/webhook")
        mock_response = MagicMock()
        mock_response.__enter__ = lambda s: s
        mock_response.__exit__ = MagicMock(return_value=False)
        mock_response.status = 500

        with patch("urllib.request.urlopen", return_value=mock_response):
            with patch("agent_observability.drift.alerts.logger") as mock_logger:
                handler(self._make_alert())
                mock_logger.warning.assert_called()

    def test_network_error_logs_exception(self) -> None:
        handler = WebhookAlertHandler("http://example.com/webhook")
        with patch("urllib.request.urlopen", side_effect=OSError("connection refused")):
            with patch("agent_observability.drift.alerts.logger") as mock_logger:
                handler(self._make_alert())
                mock_logger.exception.assert_called()

    def test_extra_headers_included(self) -> None:
        handler = WebhookAlertHandler(
            "http://example.com/webhook",
            extra_headers={"Authorization": "Bearer token123"},
        )
        mock_response = MagicMock()
        mock_response.__enter__ = lambda s: s
        mock_response.__exit__ = MagicMock(return_value=False)
        mock_response.status = 200

        with patch("urllib.request.urlopen", return_value=mock_response):
            with patch("urllib.request.Request") as mock_request:
                mock_request.return_value = MagicMock()
                handler(self._make_alert())
                call_kwargs = mock_request.call_args
                headers_passed = call_kwargs[1]["headers"] if call_kwargs[1] else call_kwargs[0][2]
                assert "Authorization" in headers_passed


# ── DriftAlertManager ─────────────────────────────────────────────────────────


class TestDriftAlertManager:
    def _no_drift_result(self) -> DriftResult:
        return _make_drift_result(drifted=False, max_z_score=0.0)

    def _high_drift_result(self, agent_id: str = "agent-1") -> DriftResult:
        return _make_drift_result(drifted=True, max_z_score=10.0, agent_id=agent_id)

    def test_no_alert_when_not_drifted(self) -> None:
        manager = DriftAlertManager(handlers=[])
        result = manager.process(self._no_drift_result())
        assert result is None

    def test_alert_fired_for_drift(self) -> None:
        fired: list[DriftAlert] = []
        manager = DriftAlertManager(handlers=[fired.append], cooldown_seconds=0.0)
        result = manager.process(self._high_drift_result())
        assert result is not None
        assert len(fired) == 1

    def test_cooldown_suppresses_repeated_alerts(self) -> None:
        fired: list[DriftAlert] = []
        manager = DriftAlertManager(handlers=[fired.append], cooldown_seconds=3600.0)
        manager.process(self._high_drift_result())
        manager.process(self._high_drift_result())
        assert len(fired) == 1

    def test_min_severity_filters_low(self) -> None:
        fired: list[DriftAlert] = []
        manager = DriftAlertManager(
            handlers=[fired.append],
            min_severity="high",
            cooldown_seconds=0.0,
        )
        # low severity result — should be filtered
        low_result = DriftResult(
            agent_id="agent-low",
            timestamp=time.time(),
            drifted=True,
            max_z_score=3.2,  # ratio 3.2/3 = 1.06 → "low" severity
            threshold=3.0,
            drifted_features={"span_count": 3.2},
            all_z_scores={"span_count": 3.2},
            baseline_age_seconds=60.0,
            window_span_count=10,
        )
        result = manager.process(low_result)
        assert result is None
        assert len(fired) == 0

    def test_add_handler(self) -> None:
        manager = DriftAlertManager(handlers=[], cooldown_seconds=0.0)
        fired: list[DriftAlert] = []
        manager.add_handler(fired.append)
        manager.process(self._high_drift_result())
        assert len(fired) == 1

    def test_handler_exception_does_not_propagate(self) -> None:
        def bad_handler(alert: DriftAlert) -> None:
            raise RuntimeError("handler failure")

        manager = DriftAlertManager(handlers=[bad_handler], cooldown_seconds=0.0)
        # Should not raise
        result = manager.process(self._high_drift_result())
        assert result is not None

    def test_different_agents_each_get_alert(self) -> None:
        fired: list[DriftAlert] = []
        manager = DriftAlertManager(handlers=[fired.append], cooldown_seconds=3600.0)
        manager.process(self._high_drift_result("agent-1"))
        manager.process(self._high_drift_result("agent-2"))
        assert len(fired) == 2

    def test_default_handler_is_console(self) -> None:
        manager = DriftAlertManager()
        assert len(manager._handlers) == 1
        assert isinstance(manager._handlers[0], ConsoleAlertHandler)


# ── Public re-exports ──────────────────────────────────────────────────────────


class TestDriftPublicImports:
    def test_public_baseline_alias(self) -> None:
        assert PublicBaseline is AgentBaseline

    def test_public_detector_alias(self) -> None:
        assert PublicDetector is DriftDetector

    def test_public_metric_detector_alias(self) -> None:
        assert PublicMetricDetector is MetricDriftDetector



"""CrewAIInstrumentor — auto-trace CrewAI task execution via monkey-patching.

Wraps ``Task.execute`` and ``Agent.execute_task`` with agent-semantic spans.
The CrewAI import is guarded; if CrewAI is not installed the instrumentor
silently becomes a no-op.
"""
from __future__ import annotations

import logging
from typing import Optional

from agent_observability.tracer.agent_tracer import AgentTracer

logger = logging.getLogger(__name__)

try:
    import crewai  # noqa: F401

    _CREWAI_AVAILABLE = True
except ImportError:
    _CREWAI_AVAILABLE = False


class CrewAIInstrumentor:
    """Instrument CrewAI with agent-semantic tracing.

    Parameters
    ----------
    tracer:
        The :class:`~agent_observability.tracer.AgentTracer` instance to use.
    """

    def __init__(self, tracer: AgentTracer) -> None:
        self._tracer = tracer
        self._original_task_execute: Optional[object] = None
        self._original_agent_execute: Optional[object] = None
        self._instrumented = False

    def instrument(self) -> None:
        """Monkey-patch CrewAI to add tracing.

        Patches:
        - ``crewai.Task.execute``
        - ``crewai.Agent.execute_task``

        Safe to call multiple times.
        """
        if self._instrumented:
            return
        if not _CREWAI_AVAILABLE:
            logger.debug("CrewAIInstrumentor: crewai not installed, skipping")
            return

        tracer = self._tracer

        try:
            from crewai import Task

            original_task_execute = Task.execute

            def _traced_task_execute(
                self_task: object,
                *args: object,
                **kwargs: object,
            ) -> object:
                task_description = getattr(self_task, "description", "crewai.task")
                agent_attr = getattr(self_task, "agent", None)
                target = getattr(agent_attr, "role", "unknown") if agent_attr else "unknown"
                with tracer.agent_delegate(target_agent=target, task=task_description[:100]) as span:
                    span.set_attribute("crewai.task.description", str(task_description)[:256])
                    return original_task_execute(self_task, *args, **kwargs)  # type: ignore[operator]

            Task.execute = _traced_task_execute  # type: ignore[method-assign]
            self._original_task_execute = original_task_execute
            logger.debug("CrewAIInstrumentor: patched Task.execute")
        except (ImportError, AttributeError) as exc:
            logger.debug("CrewAIInstrumentor: could not patch Task: %s", exc)

        try:
            from crewai import Agent

            original_agent_execute = Agent.execute_task

            def _traced_agent_execute(
                self_agent: object,
                task: object,
                *args: object,
                **kwargs: object,
            ) -> object:
                agent_role = getattr(self_agent, "role", "unknown_agent")
                with tracer.reasoning_step(step_name="execute_task", description=agent_role) as span:
                    span.set_attribute("crewai.agent.role", agent_role)
                    return original_agent_execute(self_agent, task, *args, **kwargs)  # type: ignore[operator]

            Agent.execute_task = _traced_agent_execute  # type: ignore[method-assign]
            self._original_agent_execute = original_agent_execute
            logger.debug("CrewAIInstrumentor: patched Agent.execute_task")
        except (ImportError, AttributeError) as exc:
            logger.debug("CrewAIInstrumentor: could not patch Agent: %s", exc)

        self._instrumented = True

    def uninstrument(self) -> None:
        """Remove all patches installed by :meth:`instrument`."""
        if not self._instrumented:
            return

        try:
            from crewai import Task

            if self._original_task_execute is not None:
                Task.execute = self._original_task_execute  # type: ignore[method-assign]
                self._original_task_execute = None
        except (ImportError, AttributeError):
            pass

        try:
            from crewai import Agent

            if self._original_agent_execute is not None:
                Agent.execute_task = self._original_agent_execute  # type: ignore[method-assign]
                self._original_agent_execute = None
        except (ImportError, AttributeError):
            pass

        self._instrumented = False
        logger.debug("CrewAIInstrumentor: uninstrumented")



"""Decision tracing — capture and query agent decision spans."""
from __future__ import annotations

from agent_observability.decisions.decision_span import (
    DecisionSpan,
    DecisionStatus,
)
from agent_observability.decisions.decision_tracker import (
    DecisionTracker,
    DecisionQuery,
    DecisionQueryResult,
)

__all__ = [
    "DecisionSpan",
    "DecisionStatus",
    "DecisionTracker",
    "DecisionQuery",
    "DecisionQueryResult",
]



"""DecisionSpan — capture a single agent decision point for observability.

Each DecisionSpan represents one meaningful decision made by an agent:
which option was chosen, what alternatives were considered, confidence
level, and a summary of the reasoning used.

Example
-------
::

    from agent_observability.decisions.decision_span import DecisionSpan

    span = DecisionSpan(
        agent_id="agent-1",
        decision_point="select_tool",
        chosen_option="web_search",
        alternatives_considered=["calculator", "file_read"],
        confidence=0.92,
        reasoning_summary="Web search best matches the information retrieval goal.",
    )
    print(span.span_id, span.timestamp_utc)
"""
from __future__ import annotations

import uuid
from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import Enum
from typing import Optional


class DecisionStatus(str, Enum):
    """Status of a decision span."""

    COMPLETED = "completed"
    OVERRIDDEN = "overridden"  # Human or higher-level agent overrode the decision
    FAILED = "failed"          # Decision process encountered an error


@dataclass
class DecisionSpan:
    """Captures a single agent decision for tracing and analysis.

    Attributes
    ----------
    agent_id:
        Identifier of the agent that made this decision.
    decision_point:
        Label describing what type of decision was made
        (e.g. ``"select_tool"``, ``"route_query"``, ``"choose_response"``).
    chosen_option:
        The option the agent selected.
    alternatives_considered:
        Other options that were evaluated but not chosen.
    confidence:
        Confidence score for the chosen option (0.0–1.0).
    reasoning_summary:
        Short human-readable explanation of why this option was chosen.
    span_id:
        Unique ID for this decision span. Auto-generated if not provided.
    trace_id:
        Optional trace ID to correlate this decision with a broader trace.
    parent_span_id:
        Optional parent span ID for nested decision hierarchies.
    agent_session_id:
        Optional session identifier for grouping decisions by session.
    status:
        Current status of this decision span.
    timestamp_utc:
        UTC timestamp when this decision was made. Auto-set if not provided.
    duration_ms:
        How long (in ms) the decision process took. None if not measured.
    metadata:
        Arbitrary additional structured data.
    """

    agent_id: str
    decision_point: str
    chosen_option: str
    alternatives_considered: list[str] = field(default_factory=list)
    confidence: float = 1.0
    reasoning_summary: str = ""
    span_id: str = field(default_factory=lambda: str(uuid.uuid4()))
    trace_id: Optional[str] = None
    parent_span_id: Optional[str] = None
    agent_session_id: Optional[str] = None
    status: DecisionStatus = DecisionStatus.COMPLETED
    timestamp_utc: datetime = field(
        default_factory=lambda: datetime.now(timezone.utc)
    )
    duration_ms: Optional[float] = None
    metadata: dict[str, object] = field(default_factory=dict)

    def __post_init__(self) -> None:
        """Validate confidence is in [0.0, 1.0]."""
        if not 0.0 <= self.confidence <= 1.0:
            raise ValueError(
                f"confidence must be between 0.0 and 1.0, got {self.confidence}"
            )

    @property
    def alternative_count(self) -> int:
        """Number of alternatives considered."""
        return len(self.alternatives_considered)

    @property
    def is_high_confidence(self) -> bool:
        """True if confidence >= 0.8."""
        return self.confidence >= 0.8

    @property
    def is_low_confidence(self) -> bool:
        """True if confidence < 0.5."""
        return self.confidence < 0.5

    def mark_overridden(self, reason: str = "") -> None:
        """Mark this decision as overridden by a higher authority."""
        self.status = DecisionStatus.OVERRIDDEN
        if reason:
            self.metadata["override_reason"] = reason

    def mark_failed(self, error: str = "") -> None:
        """Mark this decision as failed."""
        self.status = DecisionStatus.FAILED
        if error:
            self.metadata["failure_reason"] = error

    def to_dict(self) -> dict[str, object]:
        """Serialise to a plain dictionary."""
        return {
            "span_id": self.span_id,
            "agent_id": self.agent_id,
            "trace_id": self.trace_id,
            "parent_span_id": self.parent_span_id,
            "agent_session_id": self.agent_session_id,
            "decision_point": self.decision_point,
            "chosen_option": self.chosen_option,
            "alternatives_considered": self.alternatives_considered,
            "alternative_count": self.alternative_count,
            "confidence": self.confidence,
            "reasoning_summary": self.reasoning_summary,
            "status": self.status.value,
            "timestamp_utc": self.timestamp_utc.isoformat(),
            "duration_ms": self.duration_ms,
            "metadata": self.metadata,
        }

    @classmethod
    def from_dict(cls, data: dict[str, object]) -> "DecisionSpan":
        """Reconstruct a DecisionSpan from a dictionary.

        Parameters
        ----------
        data:
            Dictionary as produced by :meth:`to_dict`.

        Returns
        -------
        DecisionSpan
            Reconstructed span.
        """
        timestamp_str = data.get("timestamp_utc", "")
        if timestamp_str:
            try:
                timestamp = datetime.fromisoformat(timestamp_str)
                if timestamp.tzinfo is None:
                    timestamp = timestamp.replace(tzinfo=timezone.utc)
            except (ValueError, TypeError):
                timestamp = datetime.now(timezone.utc)
        else:
            timestamp = datetime.now(timezone.utc)

        status_str = data.get("status", "completed")
        try:
            status = DecisionStatus(status_str)
        except ValueError:
            status = DecisionStatus.COMPLETED

        return cls(
            agent_id=str(data.get("agent_id", "")),
            decision_point=str(data.get("decision_point", "")),
            chosen_option=str(data.get("chosen_option", "")),
            alternatives_considered=list(data.get("alternatives_considered", [])),
            confidence=float(data.get("confidence", 1.0)),
            reasoning_summary=str(data.get("reasoning_summary", "")),
            span_id=str(data.get("span_id", str(uuid.uuid4()))),
            trace_id=data.get("trace_id"),
            parent_span_id=data.get("parent_span_id"),
            agent_session_id=data.get("agent_session_id"),
            status=status,
            timestamp_utc=timestamp,
            duration_ms=data.get("duration_ms"),
            metadata=dict(data.get("metadata", {})),
        )



"""Comprehensive tests for the privacy subsystem.

Covers:
- privacy.patterns  (PiiPattern, COMMON_PATTERNS, US_PATTERNS, EU_PATTERNS,
                     get_patterns_for_jurisdiction)
- privacy.config    (CustomPatternConfig, PrivacyConfig)
- privacy.detector  (PiiMatch, DetectionResult, PiiDetector)
- privacy.redactor  (PiiRedactor, PiiRedactingSpanProcessor)
"""
from __future__ import annotations

import pytest

from agent_observability.privacy.patterns import (
    PiiPattern,
    COMMON_PATTERNS,
    EU_PATTERNS,
    US_PATTERNS,
    ALL_PATTERNS,
    PATTERNS_BY_JURISDICTION,
    get_patterns_for_jurisdiction,
)
from agent_observability.privacy.config import CustomPatternConfig, PrivacyConfig
from agent_observability.privacy.detector import (
    DetectionResult,
    PiiDetector,
    PiiMatch,
)
from agent_observability.privacy.redactor import PiiRedactor


# ── PiiPattern ─────────────────────────────────────────────────────────────────


class TestPiiPattern:
    def test_compiled_returns_pattern(self) -> None:
        pattern = PiiPattern(
            name="test",
            pattern=r"\d{4}",
            jurisdiction="common",
            replacement_token="[REDACTED_TEST]",
        )
        compiled = pattern.compiled()
        assert compiled.search("1234") is not None
        assert compiled.search("abc") is None

    def test_compiled_is_case_insensitive(self) -> None:
        pattern = PiiPattern(
            name="test",
            pattern=r"hello",
            jurisdiction="common",
            replacement_token="[REDACTED]",
        )
        assert pattern.compiled().search("HELLO") is not None


# ── Pattern collections ────────────────────────────────────────────────────────


class TestPatternCollections:
    def test_common_patterns_not_empty(self) -> None:
        assert len(COMMON_PATTERNS) > 0

    def test_us_patterns_not_empty(self) -> None:
        assert len(US_PATTERNS) > 0

    def test_eu_patterns_not_empty(self) -> None:
        assert len(EU_PATTERNS) > 0

    def test_all_patterns_combines_all(self) -> None:
        assert len(ALL_PATTERNS) == len(COMMON_PATTERNS) + len(US_PATTERNS) + len(EU_PATTERNS)

    def test_patterns_by_jurisdiction_keys(self) -> None:
        assert set(PATTERNS_BY_JURISDICTION.keys()) == {"common", "us", "eu", "all"}


class TestGetPatternsForJurisdiction:
    def test_common_jurisdiction(self) -> None:
        patterns = get_patterns_for_jurisdiction("common")
        assert patterns == COMMON_PATTERNS

    def test_us_jurisdiction_includes_common(self) -> None:
        patterns = get_patterns_for_jurisdiction("us")
        for p in COMMON_PATTERNS:
            assert p in patterns

    def test_eu_jurisdiction_includes_common(self) -> None:
        patterns = get_patterns_for_jurisdiction("eu")
        for p in COMMON_PATTERNS:
            assert p in patterns

    def test_all_jurisdiction(self) -> None:
        patterns = get_patterns_for_jurisdiction("all")
        assert patterns == ALL_PATTERNS

    def test_unknown_jurisdiction_raises(self) -> None:
        with pytest.raises(ValueError, match="Unknown jurisdiction"):
            get_patterns_for_jurisdiction("mars")


# ── Pattern detection accuracy ─────────────────────────────────────────────────


class TestCommonPatternDetection:
    def test_email_detected(self) -> None:
        detector = PiiDetector(jurisdiction="common")
        assert detector.has_pii("email me at user@example.com please")

    def test_ipv4_detected(self) -> None:
        detector = PiiDetector(jurisdiction="common")
        assert detector.has_pii("server at 192.168.1.100")

    def test_bearer_token_detected(self) -> None:
        detector = PiiDetector(jurisdiction="common")
        assert detector.has_pii("Authorization: Bearer eyJhbGciOiJIUzI1NiJ9abcdefghijklmno")

    def test_clean_text_not_detected(self) -> None:
        detector = PiiDetector(jurisdiction="common")
        assert not detector.has_pii("the quick brown fox jumps over the lazy dog")


class TestUsPatternDetection:
    def test_ssn_detected(self) -> None:
        detector = PiiDetector(jurisdiction="us")
        assert detector.has_pii("ssn: 123-45-6789")

    def test_credit_card_detected(self) -> None:
        detector = PiiDetector(jurisdiction="us")
        # Valid Visa test card
        assert detector.has_pii("card: 4111111111111111")


# ── PrivacyConfig ──────────────────────────────────────────────────────────────


class TestPrivacyConfig:
    def test_defaults(self) -> None:
        config = PrivacyConfig()
        assert config.enabled is True
        assert config.jurisdiction == "all"
        assert config.redact_span_attributes is True
        assert config.redact_log_messages is False
        assert config.custom_patterns == []
        assert "agent.id" in config.exclude_attribute_keys
        assert config.full_logging_opt_in is False
        assert "{NAME}" in config.redaction_token_template

    def test_effective_redaction_enabled_when_enabled(self) -> None:
        config = PrivacyConfig(enabled=True, full_logging_opt_in=False)
        assert config.effective_redaction_enabled() is True

    def test_effective_redaction_disabled_when_disabled(self) -> None:
        config = PrivacyConfig(enabled=False)
        assert config.effective_redaction_enabled() is False

    def test_effective_redaction_disabled_by_opt_in(self) -> None:
        config = PrivacyConfig(enabled=True, full_logging_opt_in=True)
        assert config.effective_redaction_enabled() is False

    def test_invalid_token_template_raises(self) -> None:
        with pytest.raises(Exception):
            PrivacyConfig(redaction_token_template="NO_PLACEHOLDER_HERE")

    def test_custom_pattern_config(self) -> None:
        cp = CustomPatternConfig(
            name="employee_id",
            pattern=r"EMP\d{6}",
            replacement_token="[REDACTED_EMP_ID]",
            jurisdiction="custom",
        )
        assert cp.name == "employee_id"
        assert cp.replacement_token == "[REDACTED_EMP_ID]"

    def test_jurisdiction_options(self) -> None:
        for jur in ("common", "us", "eu", "all"):
            config = PrivacyConfig(jurisdiction=jur)
            assert config.jurisdiction == jur


# ── PiiMatch / DetectionResult ─────────────────────────────────────────────────


class TestDetectionResult:
    def test_has_pii_false_when_no_matches(self) -> None:
        result = DetectionResult(original="clean text", matches=[])
        assert result.has_pii is False

    def test_has_pii_true_when_matches(self) -> None:
        match = PiiMatch(
            pattern_name="email",
            replacement_token="[REDACTED_EMAIL]",
            start=0,
            end=5,
            matched_text="test@x.com",
        )
        result = DetectionResult(original="test@x.com", matches=[match])
        assert result.has_pii is True

    def test_pattern_names(self) -> None:
        matches = [
            PiiMatch("email", "[REDACTED_EMAIL]", 0, 10, "test@x.com"),
            PiiMatch("ssn", "[REDACTED_SSN]", 20, 31, "123-45-6789"),
        ]
        result = DetectionResult(original="...", matches=matches)
        assert result.pattern_names == ["email", "ssn"]


# ── PiiDetector ────────────────────────────────────────────────────────────────


class TestPiiDetector:
    def test_detect_email(self) -> None:
        detector = PiiDetector(jurisdiction="common")
        result = detector.detect("contact user@example.com for info")
        assert result.has_pii
        assert any(m.pattern_name == "email" for m in result.matches)

    def test_detect_no_pii(self) -> None:
        detector = PiiDetector(jurisdiction="common")
        result = detector.detect("nothing to see here")
        assert not result.has_pii

    def test_detect_sorts_matches_by_position(self) -> None:
        detector = PiiDetector(jurisdiction="common")
        text = "email a@b.com and ip 192.168.1.1"
        result = detector.detect(text)
        if len(result.matches) >= 2:
            for i in range(len(result.matches) - 1):
                assert result.matches[i].start <= result.matches[i + 1].start

    def test_has_pii_fast_path(self) -> None:
        detector = PiiDetector(jurisdiction="common")
        assert detector.has_pii("a@b.com")
        assert not detector.has_pii("nothing sensitive")

    def test_scan_dict_returns_pii_keys_only(self) -> None:
        detector = PiiDetector(jurisdiction="common")
        data = {
            "user_email": "user@example.com",
            "safe_field": "nothing here",
            "count": "42",
        }
        results = detector.scan_dict(data)
        assert "user_email" in results
        assert "safe_field" not in results

    def test_scan_dict_excludes_keys(self) -> None:
        detector = PiiDetector(jurisdiction="common")
        data = {"email": "user@example.com", "trace_id": "user@example.com"}
        results = detector.scan_dict(data, exclude_keys=["trace_id"])
        assert "email" in results
        assert "trace_id" not in results

    def test_scan_dict_ignores_non_strings(self) -> None:
        detector = PiiDetector(jurisdiction="common")
        data = {"count": 42, "items": [1, 2, 3]}  # type: ignore[dict-item]
        results = detector.scan_dict(data)
        assert results == {}

    def test_extra_patterns_applied(self) -> None:
        custom = PiiPattern(
            name="employee_id",
            pattern=r"EMP\d{6}",
            jurisdiction="custom",
            replacement_token="[REDACTED_EMP]",
        )
        detector = PiiDetector(jurisdiction="common", extra_patterns=[custom])
        assert detector.has_pii("employee EMP123456 handled this")

    def test_dedup_by_name_last_writer_wins(self) -> None:
        p1 = PiiPattern("dupe", r"aaa", "common", "[A]")
        p2 = PiiPattern("dupe", r"bbb", "custom", "[B]")
        detector = PiiDetector(jurisdiction="common", extra_patterns=[p1, p2])
        # Only one pattern named "dupe" should remain
        pattern_names = [p.name for p, _ in detector._patterns]
        assert pattern_names.count("dupe") == 1


# ── PiiRedactor ────────────────────────────────────────────────────────────────


class TestPiiRedactor:
    def test_redact_email(self) -> None:
        redactor = PiiRedactor()
        redacted, count = redactor.redact("Email me at user@example.com please")
        assert "user@example.com" not in redacted
        assert count == 1

    def test_redact_no_pii(self) -> None:
        redactor = PiiRedactor()
        redacted, count = redactor.redact("nothing sensitive here")
        assert redacted == "nothing sensitive here"
        assert count == 0

    def test_redact_with_custom_template(self) -> None:
        config = PrivacyConfig(redaction_token_template="<<{NAME}>>")
        redactor = PiiRedactor(config=config)
        redacted, count = redactor.redact("contact user@example.com")
        assert "<<EMAIL>>" in redacted

    def test_redact_disabled_by_config(self) -> None:
        config = PrivacyConfig(enabled=False)
        redactor = PiiRedactor(config=config)
        text = "user@example.com"
        redacted, count = redactor.redact(text)
        assert redacted == text
        assert count == 0

    def test_redact_disabled_by_opt_in(self) -> None:
        config = PrivacyConfig(full_logging_opt_in=True)
        redactor = PiiRedactor(config=config)
        text = "user@example.com"
        redacted, count = redactor.redact(text)
        assert redacted == text

    def test_redact_dict_strings_only(self) -> None:
        redactor = PiiRedactor()
        data = {
            "email": "user@example.com",
            "count": 42,
            "nested": {"inner": "safe"},
        }
        redacted_dict, total = redactor.redact_dict(data)
        assert "user@example.com" not in str(redacted_dict["email"])
        assert redacted_dict["count"] == 42

    def test_redact_dict_excludes_config_keys(self) -> None:
        config = PrivacyConfig()
        redactor = PiiRedactor(config=config)
        data = {
            "agent.id": "user@example.com",  # should be excluded
            "message": "contact user@example.com",
        }
        redacted_dict, _ = redactor.redact_dict(data)
        assert redacted_dict["agent.id"] == "user@example.com"

    def test_redact_dict_extra_exclude_keys(self) -> None:
        redactor = PiiRedactor()
        data = {"email": "user@example.com", "skip_me": "user@example.com"}
        redacted_dict, _ = redactor.redact_dict(data, exclude_keys=["skip_me"])
        assert redacted_dict["skip_me"] == "user@example.com"

    def test_redact_dict_disabled(self) -> None:
        config = PrivacyConfig(enabled=False)
        redactor = PiiRedactor(config=config)
        data = {"email": "user@example.com"}
        redacted_dict, count = redactor.redact_dict(data)
        assert redacted_dict["email"] == "user@example.com"
        assert count == 0

    def test_custom_pattern_applied(self) -> None:
        custom_pattern = CustomPatternConfig(
            name="ticket_id",
            pattern=r"TICKET-[A-Z]{6}",
            replacement_token="[REDACTED_TICKET]",
        )
        config = PrivacyConfig(custom_patterns=[custom_pattern])
        redactor = PiiRedactor(config=config)
        redacted, count = redactor.redact("Issue TICKET-ABCDEF opened")
        assert "TICKET-ABCDEF" not in redacted
        assert count >= 1

    def test_multiple_pii_items_in_one_string(self) -> None:
        redactor = PiiRedactor()
        text = "from user@example.com and also user2@example.org"
        redacted, count = redactor.redact(text)
        assert count >= 2
        assert "user@example.com" not in redacted


# ── PiiRedactingSpanProcessor ─────────────────────────────────────────────────


class TestPiiRedactingSpanProcessor:
    def test_on_start_is_noop(self) -> None:
        from agent_observability.privacy.redactor import PiiRedactingSpanProcessor

        processor = PiiRedactingSpanProcessor()
        # Should not raise
        processor.on_start(object())

    def test_on_end_without_otel_sdk(self) -> None:
        from agent_observability.privacy.redactor import PiiRedactingSpanProcessor

        processor = PiiRedactingSpanProcessor()
        with pytest.MonkeyPatch().context() as mp:
            mp.setattr(
                "agent_observability.privacy.redactor._OTEL_SDK_AVAILABLE", False
            )
            # Should not raise
            processor.on_end(MagicMock())

    def test_shutdown_is_noop(self) -> None:
        from agent_observability.privacy.redactor import PiiRedactingSpanProcessor

        processor = PiiRedactingSpanProcessor()
        processor.shutdown()  # should not raise

    def test_force_flush_returns_true(self) -> None:
        from agent_observability.privacy.redactor import PiiRedactingSpanProcessor

        processor = PiiRedactingSpanProcessor()
        assert processor.force_flush() is True

    def test_on_end_with_mock_span_and_pii(self) -> None:
        from unittest.mock import MagicMock, patch
        from agent_observability.privacy.redactor import PiiRedactingSpanProcessor

        processor = PiiRedactingSpanProcessor()
        mock_span = MagicMock()
        mock_span.attributes = {"message": "contact user@example.com"}

        with patch("agent_observability.privacy.redactor._OTEL_SDK_AVAILABLE", True):
            processor.on_end(mock_span)
            mock_span.set_attribute.assert_called()


try:
    from unittest.mock import MagicMock
except ImportError:
    pass



"""GenericInstrumentor — decorator-based instrumentation for any Python function.

Wraps arbitrary sync and async callables with agent-semantic spans so that
any code path can be traced without coupling to a specific framework.
"""
from __future__ import annotations

import asyncio
import functools
import logging
import time
from typing import Callable, Optional, TypeVar, overload

from agent_observability.spans.types import AgentSpan, AgentSpanKind, AgentTracer

logger = logging.getLogger(__name__)

_FuncT = TypeVar("_FuncT", bound=Callable[..., object])


class GenericInstrumentor:
    """Provide decorator-based instrumentation for arbitrary functions.

    Parameters
    ----------
    tracer:
        A configured :class:`~agent_observability.spans.types.AgentTracer`.
        A default no-op tracer is used if not supplied.
    """

    def __init__(self, tracer: Optional[AgentTracer] = None) -> None:
        self._tracer = tracer or AgentTracer()

    # ── Decorators ────────────────────────────────────────────────────────────

    def instrument(
        self,
        name: str = "",
        kind: AgentSpanKind = AgentSpanKind.REASONING_STEP,
        capture_args: bool = False,
        capture_result: bool = False,
    ) -> Callable[[_FuncT], _FuncT]:
        """Return a decorator that wraps a function with an agent span.

        Parameters
        ----------
        name:
            Span name.  Defaults to the function's qualified name.
        kind:
            The :class:`AgentSpanKind` to assign to the span.
        capture_args:
            When ``True``, record a hash of the arguments as a span attribute.
        capture_result:
            When ``True``, record whether the result was truthy as a span attribute.
        """

        def decorator(func: _FuncT) -> _FuncT:
            span_name = name or f"{func.__module__}.{func.__qualname__}"

            if asyncio.iscoroutinefunction(func):
                @functools.wraps(func)
                async def async_wrapper(*args: object, **kwargs: object) -> object:
                    with self._tracer._managed_span(span_name, kind) as span:
                        if capture_args:
                            _safe_set(span, "fn.args_hash", str(hash(str(args) + str(kwargs)))[:16])
                        try:
                            result = await func(*args, **kwargs)  # type: ignore[misc]
                            if capture_result:
                                _safe_set(span, "fn.result_truthy", bool(result))
                            return result
                        except Exception as exc:
                            span.set_error(
                                error_type=type(exc).__name__,
                                recoverable=True,
                                exception=exc,
                            )
                            raise

                return async_wrapper  # type: ignore[return-value]

            else:
                @functools.wraps(func)
                def sync_wrapper(*args: object, **kwargs: object) -> object:
                    with self._tracer._managed_span(span_name, kind) as span:
                        if capture_args:
                            _safe_set(span, "fn.args_hash", str(hash(str(args) + str(kwargs)))[:16])
                        try:
                            result = func(*args, **kwargs)  # type: ignore[operator]
                            if capture_result:
                                _safe_set(span, "fn.result_truthy", bool(result))
                            return result
                        except Exception as exc:
                            span.set_error(
                                error_type=type(exc).__name__,
                                recoverable=True,
                                exception=exc,
                            )
                            raise

                return sync_wrapper  # type: ignore[return-value]

        return decorator

    def llm_call(self, name: str = "", model: str = "", provider: str = "") -> Callable[[_FuncT], _FuncT]:
        """Shortcut decorator for wrapping an LLM call function."""
        def decorator(func: _FuncT) -> _FuncT:
            span_name = name or f"{func.__qualname__}"

            @functools.wraps(func)
            def wrapper(*args: object, **kwargs: object) -> object:
                with self._tracer.llm_call(model=model, provider=provider, name=span_name) as _:
                    return func(*args, **kwargs)  # type: ignore[operator]

            return wrapper  # type: ignore[return-value]

        return decorator

    def tool_invoke(self, tool_name: str, name: str = "") -> Callable[[_FuncT], _FuncT]:
        """Shortcut decorator for wrapping a tool invocation function."""
        def decorator(func: _FuncT) -> _FuncT:
            span_name = name or func.__qualname__

            @functools.wraps(func)
            def wrapper(*args: object, **kwargs: object) -> object:
                with self._tracer.tool_invoke(tool_name=tool_name, name=span_name) as span:
                    try:
                        result = func(*args, **kwargs)  # type: ignore[operator]
                        span.set_tool(tool_name, success=True)
                        return result
                    except Exception as exc:
                        span.set_tool(tool_name, success=False, error_type=type(exc).__name__)
                        raise

            return wrapper  # type: ignore[return-value]

        return decorator


def _safe_set(span: AgentSpan, key: str, value: object) -> None:
    """Set a span attribute, ignoring type errors for non-primitive values."""
    if isinstance(value, (str, int, float, bool)):
        span.set_attribute(key, value)
    else:
        span.set_attribute(key, str(value))
